# RCAIDE/Library/Attributes/Coolants/Glycol_Water.py
# 
#
# Created:  Mar 2024, M. Clarke

# ---------------------------------------------------------------------------------------------------------------------- 
#  Imports
# ---------------------------------------------------------------------------------------------------------------------- 
from .Coolant import Coolant

# ---------------------------------------------------------------------------------------------------------------------- 
#  Glycol_Water
# ----------------------------------------------------------------------------------------------------------------------  
class Glycol_Water(Coolant):
    """Generic class of ethelyne glycol-water mixture coolant.
    """

    def __defaults__(self):
        """This sets the default values.
    
        Assumptions:
            Mixture is 50% water-50% ethylene-glycol
        
        Source: 
            Engineering Toolbox: https://www.engineeringtoolbox.com/ethylene-glycol-d_146.html
            University of Waterloo: http://www.mhtl.uwaterloo.ca/old/onlinetools/airprop/airprop.html 
        """ 
        self.tag                       = 'Glycol_Water'
        self.percent_glycol            = 0.5 
        self.density                   = 1075                       # kg/m^3
        self.specific_heat_capacity    = 3300                       # J/kg.K
        self.thermal_conductivity      = 0.387                      # W/m.K
        self.dynamic_viscosity         = 0.0019                     # Pa.s
        self.Prandtl_number            = self.specific_heat_capacity * self.dynamic_viscosity / self.thermal_conductivity
        self.kinematic_viscosity       = self.dynamic_viscosity / self.density

    def compute_cp(self,T=300):
        """ Computes the specfic heat capacity of water-glycol at a specfic temperature  
        
        Assumptions:
            Mixture is 50% water-50% glycol
            
        Source:
            Engineering Toolbox: https://www.engineeringtoolbox.com/ethylene-glycol-d_146.html
    
        Args:
            self      : glycol-water coolant  [unitless]
            T (float) : temperature           [K]
            
        Returns:
            cp (float): specfic heat capacity [J/(kg K)]         
        """ 
        return self.specific_heat_capacity
    
    def compute_absolute_viscosity(self,T=300.,p=101325.):
        """ Computes the absolute viscosity of water-glycol at a specfic temperature and pressure 
        
        Assumptions:
            Mixture is 50% water-50% glycol
            
        Source:
            Engineering Toolbox: https://www.engineeringtoolbox.com/ethylene-glycol-d_146.html
    
        Args:
            self      : glycol-water coolant  [unitless]
            T (float) : temperature           [K]
            P (float) : pressure              [Pa]
            
        Returns:
            mu (float): absolute viscosity    [kg/(m-s)] 
        """ 
        return  self.dynamic_viscosity
    
    def compute_density(self,T=300.,p=101325.):  
        """ Computes the density of water-glycol at a specfic temperature and pressure 
        
        Assumptions:
            Mixture is 50% water-50% glycol
            
        Source:
            Engineering Toolbox: https://www.engineeringtoolbox.com/ethylene-glycol-d_146.html
    
        Args:
            self       : glycol-water coolant  [unitless]
            T (float)  : temperature           [K]
            P (float)  : pressure              [Pa]
            
        Returns:
            rho (float): density               [kg/m^3]         
        """         
        return self.density  
    
    def compute_thermal_conductivity(self,T=300.,p=101325.): 
        """ Computes the density of water-glycol at a specfic temperature and pressure 
        
        Assumptions:
            Mixture is 50% water-50% glycol
            
        Source:
            University of Waterrloo: http://www.mhtl.uwaterloo.ca/old/onlinetools/airprop/airprop.html
    
        Args:
            self       : glycol-water coolant  [unitless]
            T (float)  : temperature           [K]
            P (float)  : pressure              [Pa]
            
        Returns:
            k (float)  : thermal conductivity  [W/m·K]         
        """          
        return self.thermal_conductivity
    
    
    def compute_prandtl_number(self,T=300.): 
        """ Computes the density of water-glycol at a specfic temperature and pressure 
        
        Assumptions:
            Mixture is 50% water-50% glycol
            
        Source:
            University of Waterrloo: http://www.mhtl.uwaterloo.ca/old/onlinetools/airprop/airprop.html
    
        Args:
            self       : glycol-water coolant  [unitless]
            T (float)  : temperature           [K] 
            
        Returns:
            Pr (float) : Prandtl number        [unitless]         
        """          
        
        Cp = self.compute_cp(T)
        mu = self.compute_absolute_viscosity(T)
        K  = self.compute_thermal_conductivity(T)
        return  mu*Cp/K      


# RCAIDE/Methods/Energy/Propulsors/ICE_Propulsor/__init__.py
# 

""" RCAIDE Package Setup
"""

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------

from .compute_ice_performance         import compute_ice_performance
from .append_ice_propeller_conditions import append_ice_propeller_conditions
from .unpack_ice_propeller_unknowns   import unpack_ice_propeller_unknowns
from .pack_ice_propeller_residuals    import pack_ice_propeller_residuals



# RCAIDE/Methods/Energy/Propulsors/Converters/Rotor/__init__.py
# 

""" RCAIDE Package Setup
"""

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------
 
from .                          import Design
from .                          import Wake
from .design_propeller          import design_propeller 
from .design_lift_rotor         import design_lift_rotor
from .design_prop_rotor         import design_prop_rotor
from .append_rotor_conditions   import append_rotor_conditions
from .compute_rotor_performance import compute_rotor_performance


# RCAIDE/Methods/Noise/Correlation_Buildup/Engine/external_plug_effect.py
# 
# 
# Created:  Jul 2023, M. Clarke  

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------
 
# Python package imports   
import numpy as np   

# ----------------------------------------------------------------------------------------------------------------------     
#  Ground Proximity Effect
# ----------------------------------------------------------------------------------------------------------------------         
def ground_proximity_effect (Velocity_mixed,sound_ambient,theta_m,engine_height,Diameter_mixed,frequency):
    """This function calculates the ground proximity effect, in decibels, and is used for full-scale 
    engine test stand.
        
    Assumptions:
        N/A

    Source:
        N/A

    Inputs:
        Velocity_mixed  [m/s]
        sound_ambient   [SPL]
        theta_m         [rad]
        engine_height   [m]
        Diameter_mixed  [m]
        frequency       [1/s]

    Outputs:
        GPROX_m         [dB]

    Properties Used:
        N/A 
    """ 
    # Ground proximity is applied only for the mixed jet component
    GPROX_m = (5*Velocity_mixed/sound_ambient)*np.exp(-(9*(theta_m/np.pi)-6.75)**2- \
        ((engine_height/Diameter_mixed)-2.5)**2)*(1+(np.sin((np.pi*engine_height*frequency/sound_ambient)-np.pi/2))**2)/ \
        (2+np.abs((engine_height*frequency/sound_ambient)-1))

    return GPROX_m


# RCAIDE/Library/Methods/Weights/Correlation_Buildups/Human_Powered/compute_wing_weight.py
# 
# 
# Created:  Sep 2024, M. Clarke 
 
# ----------------------------------------------------------------------------------------------------------------------
#  Wing Weight 
# ----------------------------------------------------------------------------------------------------------------------
def compute_wing_weight(Sw,bw,cw,Nwr,t_cw,Nwer,nult,GW):
    """ Compute weight of human-powered aircraft wing 
    
    Assumptions:
       All of this is from AIAA 89-2048, units are in kg. These weight estimates
       are from the MIT Daedalus and are valid for very lightweight
       carbon fiber composite structures. This may need to be solved iteratively since
       gross weight is an input.
       
    Source: 
        MIT Daedalus
        
    Inputs:
        Sw -       wing area                                                       [meters**2]
        bw -       wing span                                                       [meters]
        cw -       average wing chord                                              [meters]
        eltaw -   average rib spacing to average chord ratio                       [dimensionless]
        Nwr -      number of wing or tail surface ribs (bw^2)/(deltaw*Sw)          [dimensionless]
        t_cw -     wing airfoil thickness to chord ratio                           [dimensionless]
        Nwer -     number of wing end ribs (2*number of individual wing panels -2) [dimensionless]
        nult -     ultimate load factor                                            [dimensionless]
        GW -       aircraft gross weight                                           [kilogram]
    
    Outputs:
        Wws -      weight of wing spar                                             [kilogram]
        Wwr -      weight of wing ribs                                             [kilogram]
        Wwer -     weight of wing end ribs                                         [kilogram]
        WwLE -     weight of wing leading edge                                     [kilogram]
        WwTE -     weight of wing trailing edge                                    [kilogram]
        Wwc -      weight of wing covering                                         [kilogram]

    Properties Used:
        N/A
    """ 
    
    deltaw = (bw**2)/(Sw*Nwr)
    
    #Wing One Wire Main Spar:
    #Wws    = (bw * (3.10e-2) + (7.56e-3) * (bw**2)) * (1.0 + (nult * GW /100.0 - 2.0) / 4.0)
    
    #Wing Cantilever Main Spar:
    Wws    = (bw * (1.17e-1) + (1.1e-2) * (bw**2)) * (1.0 + (nult * GW /100.0 - 2.0) / 4.0)    
    
    #Wing Secondary Structure:
    Wwr    = Nwr * ((cw**2) * t_cw * 5.50e-2 + cw * 1.91e-3)
    Wwer   = Nwer * ((cw**2) * t_cw * 6.62e-1 + cw * 6.57e-3)
    WwLE   = 0.456 * ((Sw**2)*(deltaw**(4./3.))/bw)
    WwTE   = bw * 2.77e-2
    Wwc    = Sw * 3.08e-2
    
    weight = Wws + Wwr + Wwer + WwLE + WwTE + Wwc
    
    return weight


# RCAIDE/Compoments/Fuselages/Tube_Fuselage.py
# 
# Created:  Mar 2024, M. Clarke 

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ---------------------------------------------------------------------------------------------------------------------- 
# RCAIDE imports     
from .Fuselage import Fuselage
from RCAIDE.Library.Methods.Weights.Moment_of_Inertia.compute_fuselage_moment_of_inertia import  compute_fuselage_moment_of_inertia
 
# ---------------------------------------------------------------------------------------------------------------------- 
#  Tube_Fuselage
# ----------------------------------------------------------------------------------------------------------------------  
class Tube_Fuselage(Fuselage):
    """
    A conventional cylindrical fuselage design commonly used in commercial and general 
    aviation aircraft.

    Attributes
    ----------
    tag : str
        Unique identifier for the fuselage component, defaults to 'tube_fuselage'

    Notes
    -----
    The tube fuselage is the most common configuration in modern aircraft design.
    
    **Major Assumptions**
    
    * Cross-sections are primarily circular or near-circular
    * Smooth transitions between different diameter sections
    
    **Definitions**

    'Tube Fuselage'
        A fuselage design with predominantly circular cross-sections, typically
        consisting of a constant-diameter center section with tapered nose and
        tail sections

    See Also
    --------
    RCAIDE.Library.Components.Fuselages.Fuselage
        Base class providing core fuselage functionality
    RCAIDE.Library.Components.Fuselages.Blended_Wing_Body_Fuselage
        Alternative fuselage design for comparison
    """
    
    def __defaults__(self):
        """
        Sets default values for the tube fuselage attributes.
        """      
        self.tag = 'tube_fuselage'
        
    def compute_moment_of_inertia(self, center_of_gravity): 
        """
        Computes the moment of inertia tensor for the tube fuselage.

        Parameters
        ----------
        center_of_gravity : list
            Reference point coordinates [x, y, z] for moment calculation

        Returns
        -------
        I : ndarray
            3x3 moment of inertia tensor in kg*m^2

        See Also
        --------
        RCAIDE.Library.Methods.Weights.Moment_of_Inertia.compute_fuselage_moment_of_inertia
            Implementation of the moment of inertia calculation
        """
        I = compute_fuselage_moment_of_inertia(self,center_of_gravity) 
        return I        
  


# RCAIDE/Library/Methods/Geometry/Planform/populate_control_sections.py
#  
# Created: Oct 2024, M. Clarke

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------

# RCAIDE imports   
from RCAIDE.Framework.Core import  Container 

# package imports
import numpy   as np  

# ----------------------------------------------------------------------------------------------------------------------
#  populate_control_sections
# ---------------------------------------------------------------------------------------------------------------------- 
def populate_control_sections(wing):
    """This function takes the control surfaces defined on a wing and appends them to wing segments
    Conditional statements are used to determine where the control surface bounds are in relation 
    to the wing segments. For example, If a control surface extends beyond a wing segment, the bounds 
    on the control surface span fraction are set to be the bounds of the wing section
    
    Assumptions: 
       None

    Source:
       None
 

    Properties Used:
    N/A
    """
    w_cs  = wing.control_surfaces  
    w_seg = wing.segments 

    for i , seg in enumerate(w_seg):
        seg.control_surfaces = Container()
    
    # loop throught the control surfaces on the wing 
    for cs in w_cs :
        sf    = np.zeros(2) # set a temporary data structure to store the span fraction bounds
        sf[0] = cs.span_fraction_start
        sf[1] = cs.span_fraction_end
        
        # loop though the segments on the wing
        for i , seg in enumerate(w_seg):
            append_CS = False
            s_sf = np.zeros(2) 
            if i == 0: # the first segment (root) cannot have any control surfaces 
                pass
            else: # the following block determines where the bounds of the control surface are in relation to the segment breaks
                # Case 1
                prev_seg    =  list(w_seg.keys())[i-1]
                current_seg =  list(w_seg.keys())[i]
                if (sf[0] < w_seg[prev_seg].percent_span_location) and (sf[1] < w_seg[current_seg].percent_span_location) and (sf[1] > w_seg[prev_seg].percent_span_location) :
                    s_sf = np.array([w_seg[prev_seg].percent_span_location,sf[1]])   
                    append_CS = True 
                
                # Case 2
                elif (sf[0] < w_seg[prev_seg].percent_span_location) and (sf[1] == w_seg[current_seg].percent_span_location):
                    s_sf = np.array([w_seg[prev_seg].percent_span_location,w_seg[current_seg].percent_span_location])       
                    append_CS = True 
                    
                # Case 3   
                elif (sf[0] < w_seg[prev_seg].percent_span_location) and (sf[1] > w_seg[current_seg].percent_span_location):
                    s_sf = np.array([w_seg[prev_seg].percent_span_location,w_seg[current_seg].percent_span_location])       
                    append_CS = True                 
                
                # Case 4 
                elif (sf[0] == w_seg[prev_seg].percent_span_location) and (sf[1] < w_seg[current_seg].percent_span_location):
                    s_sf = np.array([w_seg[prev_seg].percent_span_location,sf[1]])   
                    append_CS = True 
                   
                # Case 5 
                elif (sf[0] == w_seg[prev_seg].percent_span_location) and (sf[1] == w_seg[current_seg].percent_span_location): 
                    s_sf = np.array([w_seg[prev_seg].percent_span_location,w_seg[current_seg].percent_span_location])       
                    append_CS = True
                    
                # Case 6 
                elif (sf[0] > w_seg[prev_seg].percent_span_location) and (sf[1] < w_seg[current_seg].percent_span_location):
                    s_sf = np.array([sf[0],sf[1]])
                    append_CS = True
                    
                # Case 7 
                elif (sf[0] > w_seg[prev_seg].percent_span_location) and (sf[1] == w_seg[current_seg].percent_span_location) :
                    s_sf = np.array([sf[0],w_seg[current_seg].percent_span_location]) 
                    append_CS = True                    
                    
                # Case 8
                elif (sf[0] > w_seg[prev_seg].percent_span_location) and (sf[1] > w_seg[current_seg].percent_span_location) and (sf[0] < w_seg[current_seg].percent_span_location):
                    s_sf = np.array([sf[0],w_seg[current_seg].percent_span_location]) 
                    append_CS = True
                    
                else: 
                    append_CS = False
                
                if append_CS == True:
                    # initialize the data structure for control surfaces , store results, and append to the correct segment 
                    control_surface = type(cs)() # control_surface takes came type as cs (Slat, Aileron, Data(for VLM), etc)                                             
                    control_surface.tag                   = cs.tag 
                    control_surface.span                  = cs.span*(s_sf[1]-s_sf[0])/(cs.span_fraction_end-cs.span_fraction_start)
                    control_surface.span_fraction_start   = s_sf[0] 
                    control_surface.span_fraction_end     = s_sf[1]         
                    control_surface.hinge_fraction        = cs.hinge_fraction
                    control_surface.chord_fraction        = cs.chord_fraction
                    control_surface.sign_duplicate        = cs.sign_duplicate
                    control_surface.deflection            = cs.deflection
                    control_surface.configuration_type    = cs.configuration_type
                    control_surface.gain                  = cs.gain                    
                    
                    # for calls from make_VLM_wings
                    if 'cs_type' in cs.keys():
                        control_surface.cs_type           = cs.cs_type
                    
                    w_seg[current_seg].control_surfaces.append(control_surface)        
                
    # returns an updated wing with control surfaces appended onto the wing segments                  
    return wing  




# optimize.py
# 
# Created:  Dec 2016, E. Botero
# Modified: Jun 2017, E. Botero
#           Mar 2020, M. Clarke
#           Apr 2020, M. Clarke

# ----------------------------------------------------------------------
#  Imports
# ----------------------------------------------------------------------

import scipy.optimize as opt
import numpy as np  

# ----------------------------------------------------------------------
#  Converge Root
# ----------------------------------------------------------------------

def converge_opt(segment):
    """Interfaces the mission to an optimization algorithm

    Assumptions:
    N/A

    Source:
    N/A

    Inputs:
    state.unknowns                     [Data]
    segment                            [Data]
    segment.algorithm                  [string]

    Outputs:
    state.unknowns                     [Any]

    Properties Used:
    N/A
    """     
    
    # pack up the array
    unknowns = segment.state.unknowns.pack_array()
    
    # Have the optimizer call the wrapper
    obj       = lambda unknowns:get_objective(unknowns,segment)   
    econ      = lambda unknowns:get_econstraints(unknowns,segment) 
    iecon     = lambda unknowns:get_ieconstraints(unknowns,segment)
    
    # Setup the bnds of the problem
    bnds = make_bnds(unknowns, segment)
    
    # Solve the problem, based on chosen algorithm
    if segment.algorithm == 'SLSQP':
        unknowns = opt.fmin_slsqp(obj,unknowns,f_eqcons=econ,f_ieqcons=iecon,bounds=bnds,iter=2000)
        
    elif segment.algorithm == 'SNOPT':
        
        # SNOPT imports
        import pyOpt
        import pyOpt.pySNOPT    
        
        # Have the optimizer call the wrapper
        obj_pyopt = lambda unknowns:get_problem_pyopt(unknowns,segment) 
    
        opt_prob = pyOpt.Optimization('RCAIDE',obj_pyopt)
        opt_prob.addObj(segment.objective)
    
        for ii in range(0,len(unknowns)):
            lbd = (bnds[ii][0])
            ubd = (bnds[ii][1])
            vartype = 'c'
            opt_prob.addVar(str(ii),vartype,lower=lbd,upper=ubd,value=unknowns[ii])  
    
        # Setup constraints
        segment_points = segment.state.numerics.number_of_control_points
        for ii in range(0,2*segment_points):
            opt_prob.addCon(str(ii), type='e', equal=0.)
        for ii in range(0,5*segment_points-1):
            opt_prob.addCon(str(ii+segment_points*2), type='i', lower=0.,upper=np.inf)
    
        print(opt_prob)
    
        snopt = pyOpt.pySNOPT.SNOPT()    
        outputs = snopt(opt_prob) 
    
        print(outputs)
        print(opt_prob.solution(0))

    return
    
# ----------------------------------------------------------------------
#  Helper Functions
# ----------------------------------------------------------------------
    
def get_objective(unknowns, segment):
    """ Runs the mission if the objective value is needed
    
        Assumptions:
        N/A
        
        Inputs:
        state.unknowns      [Data]
    
        Outputs:
        objective           [float]

        Properties Used:
        N/A
                                
    """      
    
    if isinstance(unknowns,np.ndarray):
        segment.state.unknowns.unpack_array(unknowns)
    else:
        segment.state.unknowns = unknowns
        
    if not np.all(segment.state.inputs_last == segment.state.unknowns):       
        segment.process.iterate(segment)
        
    objective = segment.state.objective_value
    
    return objective

def get_econstraints(unknowns, segment):
    """ Runs the mission if the equality constraint values are needed
    
        Assumptions:
        N/A
        
        Inputs:
        state.unknowns      [Data]
            
        Outputs:
        constraints          [array]

        Properties Used:
        N/A
                                
    """       
    
    if isinstance(unknowns,np.ndarray):
        segment.state.unknowns.unpack_array(unknowns)
    else:
        segment.state.unknowns = unknowns
        
    if not np.all(segment.state.inputs_last == segment.state.unknowns):       
        segment.process.iterate(segment)

    constraints = segment.state.constraint_values
    
    return constraints

def make_bnds(unknowns, segment):
    """ Automatically sets the bounds of the optimization.
    
        Assumptions:
        Restricts throttle to between 0 and 100%
        Restricts body angle from 0 to pi/2 radians
        Restricts flight path angle from 0 to pi/2 radians
        
        Inputs:
        none
            
        Outputs:
        bnds

        Properties Used:
        N/A
                                
    """      

    ones    = segment.state.ones_row(1)
    ones_m1 = segment.state.ones_row_m1(1).resize(segment.state._size)
    ones_m2 = segment.state.ones_row_m2(1).resize(segment.state._size)
    
    throttle_bnds = ones*(0.,1.)
    body_angle    = ones*(0., np.pi/2.)
    gamma         = ones*(0., np.pi/2.)
    
    if segment.air_speed_end is None:
        vels      = ones_m1*(0.,2000.)
    elif segment.air_speed_end is not None:    
        vels      = ones_m2*(0.,2000.)
    
    bnds = np.vstack([throttle_bnds,gamma,body_angle,vels])
    
    bnds = list(map(tuple, bnds))
    
    return bnds

def get_ieconstraints(unknowns, segment):
    """ Runs the mission if the inequality constraint values are needed, these are specific to a climb
    
        Assumptions:
        Time only goes forward
        CL is less than a specified limit
        CL is greater than zero
        All altitudes are greater than zero
        The vehicle accelerates not decelerates
        
        Inputs:
        state.unknowns      [Data]
            
        Outputs:
        constraints          [array]

        Properties Used:
        N/A
                                
    """      

    if isinstance(unknowns,np.ndarray):
        segment.state.unknowns.unpack_array(unknowns)
    else:
        segment.state.unknowns = unknowns
        
    if not np.all(segment.state.inputs_last == segment.state.unknowns):
        segment.process.iterate(segment)
    
    # Time goes forward, not backward
    t_final = segment.state.conditions.frames.inertial.time[-1,0]
    time_con = (segment.state.conditions.frames.inertial.time[1:,0] - segment.state.conditions.frames.inertial.time[0:-1,0])/t_final
    
    # Less than a specified CL limit
    lift_coefficient_limit = segment.lift_coefficient_limit
    CL_con = (lift_coefficient_limit  - segment.state.conditions.aerodynamics.coefficients.lift.total[:,0])/lift_coefficient_limit
    
    CL_con2   = segment.state.conditions.aerodynamics.coefficients.lift.total[:,0]
    
    # Altitudes are greater than 0
    alt_con = segment.state.conditions.freestream.altitude[:,0]/segment.altitude_end
    
    # Acceleration constraint, go faster not slower
    acc_con   = segment.state.conditions.frames.inertial.acceleration_vector[:,0]
    
    constraints = np.concatenate((time_con,CL_con,CL_con2,alt_con,acc_con))
    
    return constraints

def get_problem_pyopt(unknowns, segment):
    """ Runs the mission and obtains the objective and all constraints. This is formatted for pyopt
    
        Assumptions:
        Time only goes forward
        CL is less than a specified limit
        All altitudes are greater than zero
        
        Inputs:
        state.unknowns      [Data]
    
        Outputs:
        obj                 [float]
        con                 [array]
        fail                [boolean]

        Properties Used:
        N/A
                                
    """       

    if isinstance(unknowns,np.ndarray):
        segment.state.unknowns.unpack_array(unknowns)
    else:
        segment.state.unknowns = unknowns
        
    if not np.all(segment.state.inputs_last == segment.state.unknowns):
        segment.process.iterate(segment)
        
    obj      = segment.state.objective_value
    
    # Time goes forward, not backward
    t_final  = segment.state.conditions.frames.inertial.time[-1,0]
    time_con = (segment.state.conditions.frames.inertial.time[1:,0] - segment.state.conditions.frames.inertial.time[0:-1,0])/t_final
    
    # Less than a specified CL limit
    lift_coefficient_limit = segment.lift_coefficient_limit 
    CL_con   = (lift_coefficient_limit  - segment.state.conditions.aerodynamics.coefficients.lift.total[:,0])/lift_coefficient_limit
    
    CL_con2   = segment.state.conditions.aerodynamics.coefficients.lift.total[:,0]
    
    # Altitudes are greater than 0
    alt_con = segment.state.conditions.freestream.altitude[:,0]/segment.altitude_end
    
    # Acceleration constraint, go faster not slower
    acc_con   = segment.state.conditions.frames.inertial.acceleration_vector[:,0]
    
    # Put the equality and inequality constraints together
    constraints = np.concatenate((segment.state.constraint_values,time_con,CL_con,alt_con,acc_con,CL_con2))
    
    
    obj   = segment.state.objective_value
    const = constraints.tolist()
    fail  = np.array(np.isnan(obj) or np.isnan(np.array(const).any())).astype(int)    
    
    return obj,const,fail


# RCAIDE/Library/Components/Propulsors/Converters/Turbine.py
# (c) Copyright 2023 Aerospace Research Community LLC
# 
# Created:  Mar 2024, M. Clarke

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ---------------------------------------------------------------------------------------------------------------------- 
 # RCAIDE imports  
from RCAIDE.Library.Components                      import Component   
from RCAIDE.Library.Methods.Propulsors.Converters.Turbine.append_turbine_conditions import append_turbine_conditions

# ---------------------------------------------------------------------------------------------------------------------- 
#  Turbine
# ----------------------------------------------------------------------------------------------------------------------  
class Turbine(Component):
    """
    A turbine component model for gas turbine and turbofan engines.

    Attributes
    ----------
    tag : str
        Identifier for the turbine. Default is 'Turbine'.
        
    mechanical_efficiency : float
        Efficiency of mechanical power transmission. Default is 1.0.
        
    polytropic_efficiency : float
        Efficiency of the expansion process accounting for losses. Default is 1.0.

    Notes
    -----
    The Turbine class models the expansion and work extraction process in a 
    turbine stage. The model includes:
    
    * Work extraction calculations
    * Pressure ratio effects
    * Temperature changes
    * Efficiency losses
    * Mechanical power transmission
    * Real gas effects

    **Major Assumptions**

    * Efficiencies do not change with varying conditions

    **Definitions**

    'Mechanical Efficiency'
        Ratio of shaft power output to gas power extraction
    'Polytropic Efficiency'
        Measure of expansion process efficiency accounting for losses

    References
    ----------
    [1] Mattingly, J. D., & Boyer, K. M. (2016). Elements of propulsion: Gas 
        turbines and rockets, second edition Jack D. Mattingly, Keith M. Boyer. 
        American Institute of Aeronautics and Astronautics.

    See Also
    --------
    RCAIDE.Library.Components.Component
    RCAIDE.Library.Methods.Propulsors.Converters.Turbine.append_turbine_conditions
    """
    
    def __defaults__(self):
        """ This sets the default values for the component to function.

        Assumptions:
            None

        Source:
            None 
        """         
        #set the default values
        self.tag                               ='Turbine'
        self.mechanical_efficiency             = 1.0
        self.polytropic_efficiency             = 1.0

    def append_operating_conditions(self,segment,propulsor): 
        propulsor_conditions =  segment.state.conditions.energy[propulsor.tag]
        append_turbine_conditions(self,segment,propulsor_conditions)
        return                            
    


# RCAIDE/Library/Methods/Weights/Correlation_Buildups/FLOPS/compute_landing_gear_weight.py
# 
# 
# Created:  Sep 2024, M. Clarke

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------

# RCAIDE
import  RCAIDE
from RCAIDE.Framework.Core    import Units ,  Data

# python imports 
import  numpy as  np
 
# ----------------------------------------------------------------------------------------------------------------------
#  Landing Gear Weight 
# ----------------------------------------------------------------------------------------------------------------------
def compute_landing_gear_weight(vehicle):
    """ Calculate the weight of the main and nose landing gear of a transport aircraft

        Assumptions:
            No fighter jet, change DFTE to 1 for a fighter jet
            Aircraft is not meant for carrier operations, change CARBAS to 1 for carrier-based aircraft

        Source:
            The Flight Optimization System Weight Estimation Method

        Inputs:
            vehicle - data dictionary with vehicle properties                   [dimensionless]
                -.networks: data dictionary containing all propulsion properties
                -.design_range: design range of aircraft                        [nmi]
                -.systems.accessories: type of aircraft (short-range, commuter
                                                        medium-range, long-range,
                                                        sst, cargo)
                -.mass_properties.max_takeoff: MTOW                              [kilograms]
                -.nacelles['nacelle']                                            [meters]
                -.wings['main_wing'].dihedral                                    [radians]
                -.fuselages['fuselage'].width: fuselage width                    [meters]
                -.fuselages['fuselage'].lengths.total: fuselage total length     [meters]


        Outputs:
            output - data dictionary with main and nose landing gear weights    [kilograms]
                    output.main, output.nose

        Properties Used:
            N/A
    """
    DFTE    = 0
    CARBAS  = 0
    if vehicle.systems.accessories == "sst":
        RFACT = 0.00009
    else:
        RFACT = 0.00004
    DESRNG  = vehicle.flight_envelope.design_range / Units.nmi  # Design range in nautical miles
    WLDG    = vehicle.mass_properties.max_takeoff / Units.lbs * (1 - RFACT * DESRNG)
    
    for wing in vehicle.wings:
        if isinstance(wing,RCAIDE.Library.Components.Wings.Main_Wing):
            main_wing = wing
    
    l_f =  0
    for fuselage in vehicle.fuselages:
        if l_f < fuselage.lengths.total:
            main_fuselage = fuselage 
        
    for network in vehicle.networks:
        for propulsor in  network.propulsors:
            if 'nacelle' in propulsor:
                nacelle =  propulsor.nacelle 
                FNAC    = nacelle.diameter / Units.ft
                DIH     = main_wing.dihedral
                YEE     = np.max(np.abs(np.array(network.origin)[:, 1])) / Units.ft
                WF      = main_fuselage.width / Units.ft
                XMLG    = 12 * FNAC + (0.26 - np.tan(DIH)) * (YEE - 6 * WF)  # length of extended main landing gear
            else:
                XMLG    = 0.75 * main_fuselage.lengths.total / Units.ft  # length of extended nose landing gear
    XNLG = 0.7 * XMLG
    WLGM = (0.0117 - 0.0012 * DFTE) * WLDG ** 0.95 * XMLG ** 0.43
    WLGN = (0.048 - 0.0080 * DFTE) * WLDG ** 0.67 * XNLG ** 0.43 * (1 + 0.8 * CARBAS)

    output      = Data()
    output.main = WLGM * Units.lbs
    output.nose = WLGN * Units.lbs
    return output



# RCAIDE/Library/Methods/Weights/Correlation_Buildups/Raymer/compute_horizontal_tail_weight.py
# 
# 
# Created:  Sep 2024, M. Clarke

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------

# RCAIDE
import RCAIDE 
from RCAIDE.Framework.Core    import Units

# python imports 
import  numpy as  np
 
# ----------------------------------------------------------------------------------------------------------------------
#  Horizontal Tail Weight 
# ----------------------------------------------------------------------------------------------------------------------
def compute_horizontal_tail_weight(vehicle, wing, elevator_fraction=0.4):
    """ Calculates horizontal tail weight based on Raymer method

        Assumptions:
            If all-moving horizontal tail, change Kuht to 1.143
        Source:
            Aircraft Design: A Conceptual Approach

        Inputs:
            vehicle - data dictionary with vehicle properties                    [dimensionless]
                -.mass_properties.max_takeoff: MTOW                             [kilograms]
                -.flight_envelope.ultimate_load: ultimate load factor (default: 3.75)
                -.wings['main_wing']: data dictionary with properties of main wing
                    -.aerodynamic_center: aerodynamic center as measured from root leading edge
                    -.origin: root of main wing as measured from nose of aircraft
                -.fuselages['fuselage'].width: width of the fuselage
            wing    - data dictionary with specific tail properties              [dimensionless]
                -.areas.reference: tail surface area                            [m^2}
                -.origin: location of tail measured from nose
                -.aerodynamic_center: location of ac measured from leading edge
                -.sweeps.quarter_chord: quarter chord sweep of tail             [rad]
                -.thickness_to_chord: t/c of tail
                -.span.projected: project span of tail                          [m]
                -.aspect_ratio: aspect ratio of wing
            elevator_fraction - fraction of horizontal tail for elevator = 0.4

        Outputs:
            tail_weight: horizontal tail weight                                [kilograms]

        Properties Used:
            N/A
    """

    ref_wing = None 
    for wing in  vehicle.wings:
        if isinstance(wing, RCAIDE.Library.Components.Wings.Main_Wing):
            ref_wing  =  wing
    
    S = 0
    if ref_wing == None:
        for wing in  vehicle.wings:
            if S < wing.areas.reference:
                ref_wing = wing
    L_fuselage = 0
    for fuselage in vehicle.fuselages:
        if L_fuselage < fuselage.lengths.total:
            ref_fuselage = fuselage
            
    Kuht    = 1 # not a all-moving unit horizontal tail
    Fw      = ref_fuselage.width / Units.ft
    Bh      = wing.spans.projected / Units.ft
    DG      = vehicle.mass_properties.max_takeoff / Units.lbs
    Sht     = wing.areas.reference / Units.ft ** 2
    Lt      = (wing.origin[0][0] + wing.aerodynamic_center[0] - ref_wing.origin[0][0] -
                ref_wing.aerodynamic_center[0]) / Units.ft
    Ky      = 0.3 * Lt
    sweep   = wing.sweeps.quarter_chord
    Ah      = wing.aspect_ratio
    Se      = elevator_fraction * Sht

    tail_weight = 0.0379 * Kuht * (1 + Fw / Bh) ** (-0.25) * DG ** 0.639 *\
                  vehicle.flight_envelope.ultimate_load ** 0.1 * Sht ** 0.75 * Lt ** -1 *\
                  Ky ** 0.704 * np.cos(sweep) ** (-1) * Ah ** 0.166 * (1 + Se / Sht) ** 0.1
    return tail_weight * Units.lbs



# RCAIDE/Library/Components/Energy/Sources/Battery_Modules/Aluminum_Air.py
# 
# 
# Created:  Mar 2024, M. Clarke 

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------

 # RCAIDE imports
from RCAIDE.Framework.Core import Units
from .Generic_Battery_Module    import Generic_Battery_Module

# ----------------------------------------------------------------------------------------------------------------------
#  Aluminum_Air
# ----------------------------------------------------------------------------------------------------------------------   
class Aluminum_Air(Generic_Battery_Module):
    """
    Class for modeling aluminum-air battery characteristics and performance
    
    Attributes
    ----------
    tag : str
        Identifier for the battery module (default: 'Aluminum Air')
        
    cell.specific_energy : float
        Energy capacity per unit mass [J/kg] (default: 1300 Wh/kg)
        
    cell.specific_power : float
        Maximum power output per unit mass [W/kg] (default: 0.2 kW/kg)
        
    mass_gain_factor : float
        Rate of mass increase during discharge [kg/Wh] (default: 0.000110145 kg/Wh)
        
    cell.water_mass_gain_factor : float
        Rate of water consumption [kg/Wh] (default: 0.000123913 kg/Wh)
        
    cell.aluminum_mass_factor : float
        Rate of aluminum consumption [kg/Wh] (default: 0.000123828 kg/Wh)
        
    cell.ragone.const_1 : float
        First constant in Ragone curve fit [kW/kg] (default: 0.8439 kW/kg)
        
    cell.ragone.const_2 : float
        Second constant in Ragone curve fit [1/(Wh/kg)] (default: -4.8647e-4)
        
    cell.ragone.lower_bound : float
        Minimum specific energy on Ragone plot [Wh/kg] (default: 1100 Wh/kg)
        
    cell.ragone.upper_bound : float
        Maximum specific energy on Ragone plot [Wh/kg] (default: 1600 Wh/kg)

    Notes
    -----
    Aluminum-air batteries are a type of metal-air battery that uses aluminum
    as the anode and ambient oxygen as the cathode. During discharge, the battery
    consumes aluminum and water while gaining mass from the reaction products.

    **Definitions**

    'Ragone Plot'
        Graph showing the relationship between specific power and specific energy,
        characterized by constants that define the performance envelope
        
    'Mass Gain Factor'
        Rate at which the battery gains mass during discharge due to the
        conversion of aluminum to aluminum hydroxide

    See Also
    --------
    RCAIDE.Library.Components.Energy.Sources.Battery_Modules.Generic_Battery_Module
        Base battery module class
    """
    
    def __defaults__(self):
        """
        Sets default values for aluminum-air battery attributes
        """
        self.tag                         = 'Aluminum Air'
        self.cell.specific_energy        = 1300.*Units.Wh/Units.kg    # convert to Joules/kg
        self.cell.specific_power         = 0.2*Units.kW/Units.kg      # convert to W/kg
        self.mass_gain_factor            = 0.000110145*Units.kg/Units.Wh
        self.cell.water_mass_gain_factor = 0.000123913*Units.kg/Units.Wh
        self.cell.aluminum_mass_factor   = 0.000123828*Units.kg/Units.Wh # aluminum consumed per energy
        self.cell.ragone.const_1         = 0.8439*Units.kW/Units.kg
        self.cell.ragone.const_2         = -4.8647e-004/(Units.Wh/Units.kg)
        self.cell.ragone.lower_bound     = 1100.*Units.Wh/Units.kg
        self.cell.ragone.upper_bound     = 1600.*Units.Wh/Units.kg


# RCAIDE/Library/Missions/Common/Pre_Process/aerodynamics.py
# 
# 
# Created:  Jul 2023, M. Clarke

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ---------------------------------------------------------------------------------------------------------------------- 
from RCAIDE.Library.Methods.Geometry.Planform  import wing_segmented_planform, wing_planform

# ----------------------------------------------------------------------------------------------------------------------
#  aerodynamics
# ----------------------------------------------------------------------------------------------------------------------  
def aerodynamics(mission):
    """ Runs aerdoynamics model and build surrogate
    
        Assumptions:
            N/A
        
        Inputs:
            None
            
        Outputs:
            None  
    """
    
        
    last_tag = None
    for tag,segment in mission.segments.items():  
        if segment.analyses.aerodynamics != None:
            # ensure all properties of wing are computed before drag calculations  
            vehicle =  segment.analyses.aerodynamics.vehicle
            for wing in  vehicle.wings:
                if len(wing.segments) > 1: 
                    wing_segmented_planform(wing)
                else:
                    wing_planform(wing)
                
            if (last_tag!=  None) and  ('compute' in mission.segments[last_tag].analyses.aerodynamics.process.keys()): 
                segment.analyses.aerodynamics.process.compute.lift.inviscid_wings = mission.segments[last_tag].analyses.aerodynamics.process.compute.lift.inviscid_wings
                segment.analyses.aerodynamics.surrogates       = mission.segments[last_tag].analyses.aerodynamics.surrogates 
                segment.analyses.aerodynamics.reference_values = mission.segments[last_tag].analyses.aerodynamics.reference_values  
            else:          
                aero   = segment.analyses.aerodynamics
                aero.initialize()   
                last_tag = tag  
    return 


# RCAIDE/Library/Missions/Segments/Hover/Descent.py
# 
# 
# Created:  Jul 2023, M. Clarke 
 
# ----------------------------------------------------------------------------------------------------------------------  
#  Initialize Conditions
# ----------------------------------------------------------------------------------------------------------------------   
def initialize_conditions(segment):
    """Sets the specified conditions which are given for the segment type.

    Assumptions:
    Descent segment with a constant rate.

    Source:
    N/A

    Inputs:
    segment.altitude_start                              [meters]
    segment.altitude_end                                [meters]
    segment.descent_rate                                [meters/second]
    segment.state.numerics.dimensionless.control_points [Unitless]
    segment.state.conditions.frames.inertial.time       [seconds]

    Outputs:
    conditions.frames.inertial.velocity_vector  [meters/second]
    conditions.frames.inertial.position_vector  [meters]
    conditions.freestream.altitude              [meters]
    conditions.frames.inertial.time             [seconds]

    Properties Used:
    N/A
    """      
    
    # unpack
    descent_rate = segment.descent_rate
    alt0         = segment.altitude_start 
    altf         = segment.altitude_end
    t_nondim     = segment.state.numerics.dimensionless.control_points
    t_initial    = segment.state.conditions.frames.inertial.time[0,0]
    conditions   = segment.state.conditions  

    # check for initial altitude
    if alt0 is None:
        if not segment.state.initials: raise AttributeError('initial altitude not set')
        alt0 = -1.0 *segment.state.initials.conditions.frames.inertial.position_vector[-1,2]

    # discretize on altitude
    alt = t_nondim * (altf-alt0) + alt0
    
    # process velocity vector
    v_z = descent_rate # z points down    
    dt  = (alt0 - altf)/descent_rate

    # rescale operators
    t = t_nondim * dt

    # pack
    t_initial = segment.state.conditions.frames.inertial.time[0,0]
    segment.state.conditions.frames.inertial.time[:,0] = t_initial + t[:,0]    
    
    # pack conditions    
    conditions.frames.inertial.velocity_vector[:,0] = 0.
    conditions.frames.inertial.velocity_vector[:,2] = v_z
    conditions.frames.inertial.position_vector[:,2] = -alt[:,0] # z points down
    conditions.freestream.altitude[:,0]             =  alt[:,0] # positive altitude in this context
    conditions.frames.inertial.time[:,0]            = t_initial + t[:,0]


# RCAIDE/Framework/Analyses/Mission/Segments/Single_Point/__init__.py
# 

"""RCAIDE Package Setup
"""

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------

from .Set_Speed_Set_Altitude               import Set_Speed_Set_Altitude
from .Set_Speed_Set_Throttle               import Set_Speed_Set_Throttle
from .Set_Speed_Set_Altitude_No_Propulsion import Set_Speed_Set_Altitude_No_Propulsion
from .Set_Speed_Set_Altitude_AVL_Trimmed   import Set_Speed_Set_Altitude_AVL_Trimmed


# RCAIDE/Library/Missions/Segments/Hover/Hover.py
# 
# 
# Created:  Jul 2023, M. Clarke 
 
# ----------------------------------------------------------------------------------------------------------------------  
#  Initialize Conditions
# ----------------------------------------------------------------------------------------------------------------------  

def initialize_conditions(segment):
    """Sets the specified conditions which are given for the segment type.

    Assumptions:
    Descent segment with a constant rate.

    Source:
    N/A

    Inputs:
    segment.altitude                            [meters]
    segment.tim                                 [second]
    state.numerics.dimensionless.control_points [Unitless]
    state.conditions.frames.inertial.time       [seconds]

    Outputs:
    conditions.frames.inertial.velocity_vector  [meters/second]
    conditions.frames.inertial.position_vector  [meters]
    conditions.freestream.altitude              [meters]
    conditions.frames.inertial.time             [seconds]

    Properties Used:
    N/A
    """       
    
    # unpack
    alt        = segment.altitude
    duration   = segment.time
    conditions = segment.state.conditions   
    
    
    # check for initial altitude
    if alt is None:
        if not segment.state.initials: raise AttributeError('altitude not set')
        alt = -1.0 *segment.state.initials.conditions.frames.inertial.position_vector[-1,2]      
    
    # dimensionalize time
    t_initial = conditions.frames.inertial.time[0,0]
    t_nondim  = segment.state.numerics.dimensionless.control_points
    time      =  t_nondim * (duration) + t_initial
    
    # pack
    segment.state.conditions.freestream.altitude[:,0]             = alt
    segment.state.conditions.frames.inertial.position_vector[:,2] = -alt # z points down
    segment.state.conditions.frames.inertial.velocity_vector[:,0] = 0.
    segment.state.conditions.frames.inertial.time[:,0]            = time[:,0]    



## @ingroup Optimization-Package_Setups
# additive_setup.py
#
# Created:  Apr 2017, T. MacDonald
# Modified: Jun 2017, T. MacDonald
#           Oct 2019, T. MacDonald
#           Jun 2020, M. Clarke

# ----------------------------------------------------------------------
#  Imports
# ----------------------------------------------------------------------

import numpy as np
import RCAIDE
try:
    import pyOpt
    import pyOpt.pySNOPT
    import pyOpt.pyALPSO
except:
    pass
import sklearn
from sklearn import gaussian_process
from RCAIDE.Framework.Optimization.Common import helper_functions as help_fun
from RCAIDE.Library.Methods.Utilities.latin_hypercube_sampling import latin_hypercube_sampling
from scipy.stats import norm
import os
import sys
from scipy.optimize import minimize
from scipy.optimize import shgo

# ----------------------------------------------------------------------
#  Additive Solve Functions
# ----------------------------------------------------------------------
class Additive_Solver():
    def __init__(self):
        # These defaults are chosen since they are built in to scipy and
        # therefore are always available when running RCAIDE
        self.local_optimizer  = 'SLSQP'
        self.global_optimizer = 'SHGO'
        return

    ## @ingroup Optimization-Package_Setups
    def Additive_Solve(self,problem,num_fidelity_levels=2,num_samples=10,max_iterations=10,
                       tolerance=1e-6,opt_type='basic',num_starts=3,print_output=True):
        """Solves a multifidelity problem using an additive corrections
    
        Assumptions:
        N/A
    
        Source:
        N/A
    
        Inputs:
        problem             [nexus()]
        num_fidelity_levels [int]
        num_samples         [int]
        max_iterations      [int]
        tolerance           [float]
        opt_type            [str]
        num_starts          [int]
        print_output        [bool]
        
        Outputs:
        (fOpt,xOpt)  [tuple]
    
        Properties Used:
        N/A
        """        
        
        if print_output == False:
            devnull = open(os.devnull,'w')
            sys.stdout = devnull    
        
        if num_fidelity_levels != 2:
            raise NotImplementedError('Additive corrections are only implemented for 2 fidelity levels.')
        
        # History writing
        f_out = open('add_hist.txt','w')
        import datetime
        f_out.write(str(datetime.datetime.now())+'\n')
        
        inp = problem.optimization_problem.inputs
        obj = problem.optimization_problem.objective
        con = problem.optimization_problem.constraints 
    
        # Set inputs
        nam = inp[:,0] # Names
        ini = inp[:,1] # Initials
        bnd = inp[:,2] # Bounds
        scl = inp[:,3] # Scale
        typ = inp[:,4] # Type
    
        (x,scaled_constraints,x_low_bound,x_up_bound,con_up_edge,con_low_edge) = self.scale_vals(inp, con, ini, bnd, scl)  
        
        # Get initial set of samples
        x_samples = latin_hypercube_sampling(len(x),num_samples,bounds=(x_low_bound,x_up_bound),criterion='center')
        
        # Initialize objective and constraint variables
        f = np.zeros([num_fidelity_levels,num_samples])
        g = np.zeros([num_fidelity_levels,num_samples,len(scaled_constraints)])
        
        for level in range(1,num_fidelity_levels+1):
            problem.fidelity_level = level
            for ii,x in enumerate(x_samples):
                res = self.evaluate_model(problem,x,scaled_constraints)
                f[level-1,ii]    = res[0]  # objective value
                g[level-1,ii,:]  = res[1]  # constraints vector
        
        converged = False
        
        for kk in range(max_iterations):
            # Build objective surrogate
            f_diff = f[1,:] - f[0,:]
            f_additive_surrogate_base = gaussian_process.GaussianProcessRegressor()
            f_additive_surrogate = f_additive_surrogate_base.fit(x_samples, f_diff)     
            
            # Build constraint surrogate
            g_diff = g[1,:] - g[0,:]
            g_additive_surrogate_base = gaussian_process.GaussianProcessRegressor()
            g_additive_surrogate = g_additive_surrogate_base.fit(x_samples, g_diff)     
            
            # Optimize corrected model
            
            # Chose method ---------------
            if opt_type == 'basic': # Next point determined by surrogate optimum
                problem.fidelity_level = 1
                x_eval = latin_hypercube_sampling(len(x),1,bounds=(x_low_bound,x_up_bound),criterion='random')[0]
                
                if self.local_optimizer == 'SNOPT':
                    opt_prob = pyOpt.Optimization('RCAIDE',self.evaluate_corrected_model, \
                                              obj_surrogate=f_additive_surrogate,cons_surrogate=g_additive_surrogate)      
                    
                    # Set up opt_prob
                    self.initialize_opt_vals(opt_prob,obj,inp,x_low_bound,x_up_bound,con_low_edge,con_up_edge,nam,con,x_eval)  
                       
                    opt = pyOpt.pySNOPT.SNOPT()      
                    
                    outputs = opt(opt_prob, sens_type='FD',problem=problem, \
                                  obj_surrogate=f_additive_surrogate,cons_surrogate=g_additive_surrogate)#, sens_step = sense_step)  
                    fOpt = outputs[0][0]
                    xOpt = outputs[1]
                
                elif self.local_optimizer == 'SLSQP':
                
                    x0,constraints = self.initialize_opt_vals_SLSQP(obj,inp,x_low_bound,x_up_bound,con_low_edge,con_up_edge,nam,con,x_eval,problem,g_additive_surrogate)
    
                    res = minimize(self.evaluate_corrected_model, x0,constraints=constraints,args=(problem,f_additive_surrogate,g_additive_surrogate),options={'ftol':1e-6,'disp':True})
                    fOpt = res['fun']
                    xOpt = res['x']
                    
                else:
                    raise NotImplementedError
    
            elif opt_type == 'MEI': # Next point determined by maximum expected improvement
                fstar = np.min(f[1,:])
                problem.fidelity_level = 1
                
                if self.global_optimizer == 'ALPSO':
                    opt_prob = pyOpt.Optimization('RCAIDE',self.evaluate_expected_improvement, \
                                              obj_surrogate=f_additive_surrogate,cons_surrogate=g_additive_surrogate,fstar=fstar)     
                    
                    # Set up opt_prob
                    self.initialize_opt_vals(opt_prob,obj,inp,x_low_bound,x_up_bound,con_low_edge,con_up_edge,nam,con,None)     
                       
                    # Use a global optimizer
                    opt = pyOpt.pyALPSO.ALPSO()    
                    opt.setOption('maxOuterIter',value=20)
                    opt.setOption('seed',value=1.)                    
                    
                    outputs = opt(opt_prob,problem=problem, \
                                  obj_surrogate=f_additive_surrogate,cons_surrogate=g_additive_surrogate,fstar=fstar,cons=con)#, sens_step = sense_step)
                    fOpt  = np.nan 
                    imOpt = outputs[0]
                    xOpt  = outputs[1]
                    
                elif self.global_optimizer == 'SHGO':
                
                    xb, shgo_cons = self.initialize_opt_vals_SHGO(obj, inp, x_low_bound, x_up_bound, con_low_edge, con_up_edge, nam, con, problem, g_additive_surrogate)
                
                    #self.global_optimizer = 'SHGO'
                    options = {} 
                    res = shgo(self.evaluate_expected_improvement, xb, iters=2, args=(problem,f_additive_surrogate,g_additive_surrogate,fstar),constraints=shgo_cons,options=options)
                   
                    
                    fOpt  = np.nan 
                    imOpt = res['fun']
                    xOpt  = res['x']    
                    
                else:
                    raise NotImplementedError
            
            # ---------------------------------
            
            complete_flag = False
            if np.any(np.isnan(xOpt)):
                complete_flag = True
            else:
            
                # Add new samples and check objective and constraint values
                f = np.hstack((f,np.zeros((num_fidelity_levels,1))))
                g = np.hstack((g,np.zeros((num_fidelity_levels,1,len(con)))))
                x_samples = np.vstack((x_samples,xOpt))
                for level in range(1,num_fidelity_levels+1):
                    problem.fidelity_level = level
                    res = self.evaluate_model(problem,xOpt,scaled_constraints)
                    f[level-1][-1] = res[0]
                    g[level-1][-1] = res[1]
                    
                # History writing
                f_out.write('Iteration: ' + str(kk+1)    + '\n')
                f_out.write('x0       : ' + str(xOpt[0]) + '\n')
                f_out.write('x1       : ' + str(xOpt[1]) + '\n')
                if opt_type == 'basic':
                    f_out.write('expd hi  : ' + str(fOpt) + '\n')
                elif opt_type == 'MEI':
                    f_out.write('expd imp : ' + str(imOpt) + '\n')
                f_out.write('low obj : ' + str(f[0][-1]) + '\n')
                f_out.write('hi  obj : ' + str(f[1][-1]) + '\n') 
            if kk == (max_iterations-1) or complete_flag == True: # Reached maximum number of iterations
                f_diff = f[1,:] - f[0,:]
                if opt_type == 'basic': # If basic setting f already has the expected optimum
                    problem.fidelity_level = 2
                    fOpt = self.evaluate_model(problem,xOpt,scaled_constraints)[0][0]
                elif opt_type == 'MEI': # If MEI, find the optimum of the final surrogate
                
                    min_ind = np.argmin(f[1])
                    x_eval = x_samples[min_ind]
                
                    if self.local_optimizer == 'SNOPT':
                        opt_prob = pyOpt.Optimization('RCAIDE',self.evaluate_corrected_model, \
                                                      obj_surrogate=f_additive_surrogate,cons_surrogate=g_additive_surrogate)                         
                        
                        # Set up opt_prob
                        self.initialize_opt_vals(opt_prob,obj,inp,x_low_bound,x_up_bound,con_low_edge,con_up_edge,nam,con,x_eval)                           
                        
                        fOpt, xOpt = self.run_objective_optimization(opt_prob,problem,f_additive_surrogate,g_additive_surrogate)
                    
                    elif self.local_optimizer == 'SLSQP':
                        problem.fidelity_level = 1
                        x0,constraints = self.initialize_opt_vals_SLSQP(obj,inp,x_low_bound,x_up_bound,con_low_edge,con_up_edge,nam,con,x_eval,problem,g_additive_surrogate)
                        
                        res = minimize(self.evaluate_corrected_model, x0,constraints=constraints,args=(problem,f_additive_surrogate,g_additive_surrogate),options={'ftol':1e-6,'disp':True})
                        fOpt = res['fun']
                        xOpt = res['x'] 
                    
                    problem.fidelity_level = 2
                    fOpt = self.evaluate_model(problem,xOpt,scaled_constraints)[0][0]               
            
                    f_out.write('x0_opt  : ' + str(xOpt[0]) + '\n')
                    f_out.write('x1_opt  : ' + str(xOpt[1]) + '\n')                
                    f_out.write('final opt : ' + str(fOpt) + '\n')
                    
                print('Iteration Limit Reached')
                break        
                
            
            if np.abs(fOpt-f[1][-1]) < tolerance: # Converged within a tolerance
                print('Convergence reached')      
                f_out.write('Convergence reached')
                f_diff = f[1,:] - f[0,:]
                converged = True
                if opt_type == 'MEI':
                    
                    problem.fidelity_level = 1
                    min_ind = np.argmin(f[1])
                    x_eval = x_samples[min_ind]
                    
                    if self.local_optimizer == 'SNOPT':
                    
                        opt_prob = pyOpt.Optimization('RCAIDE',self.evaluate_corrected_model, \
                                                      obj_surrogate=f_additive_surrogate,cons_surrogate=g_additive_surrogate)       
                    
                        initalize_opt_vals(opt_prob,obj,inp,x_low_bound,x_up_bound,con_low_edge,con_up_edge,nam,con,x_eval)    
                    
                        opt = pyOpt.pySNOPT.SNOPT()      
                    
                        
                        outputs = opt(opt_prob, sens_type='FD',problem=problem, \
                                      obj_surrogate=f_additive_surrogate,cons_surrogate=g_additive_surrogate)#, sens_step = sense_step)  
                        fOpt = outputs[0][0]
                        xOpt = outputs[1]
                        
                    elif self.local_optimizer == 'SLSQP':
                    
                        x0,constraints = self.initialize_opt_vals_SLSQP(opt_prob,obj,inp,x_low_bound,x_up_bound,con_low_edge,con_up_edge,nam,con,x_eval,problem,g_additive_surrogate)
                        
                        res = minimize(self.evaluate_corrected_model, x0,constraints=constraints,args=(problem,f_additive_surrogate,g_additive_surrogate),options={'ftol':1e-6,'disp':True})
                        fOpt = res['fun']
                        xOpt = res['x']
                        
                    else:
                        raise NotImplementedError
                    
                    problem.fidelity_level = 2
                    fOpt = self.evaluate_model(problem,xOpt,scaled_constraints)[0][0]                      
                    
                    f_out.write('x0_opt  : ' + str(xOpt[0]) + '\n')
                    f_out.write('x1_opt  : ' + str(xOpt[1]) + '\n')                
                    f_out.write('final opt : ' + str(fOpt) + '\n')            
                break        
            
            fOpt = f[1][-1]*1.
        
        if converged == False:
            print('Iteration Limit reached')
            f_out.write('Maximum iteration limit reached')
        
        # Save sample data
        np.save('x_samples.npy',x_samples)
        np.save('f_data.npy',f)
        f_out.close()
        print(fOpt,xOpt)
        if print_output == False:
            sys.stdout = sys.__stdout__
        
        # Format objective function to array, ensure output consistency
        if np.isscalar(fOpt):
            FOpt = np.array([fOpt])
        else:
            FOpt = fOpt.astype(np.double)
        
        return (FOpt,xOpt)
        
    ## @ingroup Optimization-Package_Setups    
    def evaluate_model(self,problem,x,cons):
        """Solves the optimization problem to get the objective and constraints
    
        Assumptions:
        N/A
    
        Source:
        N/A
    
        Inputs:
        problem   [nexus()]
        x         [array]
        cons      [array]
        
        Outputs:
        f         [float]
        g         [array]
    
        Properties Used:
        N/A    
        """
        f  = np.array(0.)
        g  = np.zeros(np.shape(cons))
        
        f  = problem.objective(x)
        g  = problem.all_constraints(x)
        
        return f,g
    
    ## @ingroup Optimization-Package_Setups    
    def evaluate_corrected_model(self,x,problem=None,obj_surrogate=None,cons_surrogate=None):
        """Evaluates the corrected model with the low fidelity plus the corrections
    
        Assumptions:
        N/A
    
        Source:
        N/A
    
        Inputs:
        x              [array]
        problem        [nexus()]
        obj_surrogate  [fun()]
        cons_surrogate [fun()]
        
        Outputs:
        obj            [float]
        const          [array]
        fail           [bool]
    
        Properties Used:
        N/A    
        
        """
        
        obj   = problem.objective(x)
        const = problem.all_constraints(x).tolist()
        fail  = np.array(np.isnan(obj.tolist()) or np.isnan(np.array(const).any())).astype(int)
        
        obj_addition  = obj_surrogate.predict(np.atleast_2d(x))
        cons_addition = cons_surrogate.predict(np.atleast_2d(x))
        
        obj   = obj + obj_addition
        const = const + cons_addition
        const = const.tolist()[0]
        
        self.const_list = const
    
        print('Inputs')
        print(x)
        print('Obj')
        print(obj)
        print('Con')
        print(const)
            
        if self.local_optimizer == 'SNOPT':
            return obj,const,fail
        elif self.local_optimizer == 'SLSQP':
            return obj
        else:
            raise NotImplementedError('Selected local optimizer is not implemented.')
    
    ## @ingroup Optimization-Package_Setups
    def evaluate_expected_improvement(self,x,problem=None,obj_surrogate=None,cons_surrogate=None,fstar=np.inf,cons=None):
        """Evaluates the expected improvement of the point x
    
        Assumptions:
        N/A
    
        Source:
        N/A
    
        Inputs:
        x              [array]
        problem        [nexus()]
        obj_surrogate  [fun()]
        cons_surrogate [fun()]
        fstar          [float]
        cons           [vector]
        
        Outputs:
        -EI            [float]
        const          [array]
        fail           [bool]
    
        Properties Used:
        N/A    
        
        """    
    
        if np.any(np.isnan(x)):
            if self.global_optimizer == 'ALPSO':
                raise ValueError('Unknown error in ALPSO optimizer created NaN values.')
            return np.inf
    
        obj   = problem.objective(x)
        const = problem.all_constraints(x).tolist()
        fail  = np.array(np.isnan(obj.tolist()) or np.isnan(np.array(const).any())).astype(int)
        
        # Get uncertainty information
        obj_addition, obj_sigma   = obj_surrogate.predict(np.atleast_2d(x),return_std=True)
        cons_addition, cons_sigma = cons_surrogate.predict(np.atleast_2d(x),return_std=True)
        
        fhat  = obj[0] + obj_addition
        # Calculate expected improvement (based on Schonlau, Computer Experiments and Global Optimization, 1997)
        EI    = (fstar-fhat)*norm.cdf((fstar-fhat)/obj_sigma) + obj_sigma*norm.pdf((fstar-fhat)/obj_sigma)
        const = const + cons_addition
        EI    = np.log(EI)
        if EI == -np.inf:
            EI = -1000
        
        if self.global_optimizer == 'ALPSO':
            # Adjust signs for optimizer (this is specific to ALPSO)
            signs  = np.ones([1,len(cons)])
            offset = np.zeros([1,len(cons)])
            for ii,con in enumerate(cons):
                if cons[ii][1] == '>':
                    signs[0,ii] = -1
                offset[0,ii] = cons[ii][2]
            
            
            const = const*signs - offset*signs
            const = const.tolist()[0]
        
            print('Inputs')
            print(x)
            print('Obj')
            print(-EI)
            print('Con')
            print(const)
                
            return -EI,const,fail
        elif self.global_optimizer == 'SHGO':
            return -EI
    
    ## @ingroup Optimization-Package_Setups    
    def scale_vals(self,inp,con,ini,bnd,scl):
        """Scales values to help setup the problem
    
        Assumptions:
        N/A
    
        Source:
        N/A
    
        Inputs:
        inp                         [array]
        con                         [array]
        ini                         [array]
        bnd                         [array]
        scl                         [array]
        
        Outputs:
            tuple:
                x                   [array]
                scaled_constraints  [array]
                x_low_bounds        [array]
                x_up_bounds         [array]
                con_up_edge         [array]
                con_low_edge        [array]
    
        Properties Used:
        N/A    
        
        """     
    
        # Pull out the constraints and scale them
        bnd_constraints = help_fun.scale_const_bnds(con)
        scaled_constraints = help_fun.scale_const_values(con,bnd_constraints)
    
        x            = ini/scl        
        x_low_bound  = []
        x_up_bound   = []
        edge         = []
        con_up_edge  = []
        con_low_edge = []
    
        for ii in range(0,len(inp)):
            x_low_bound.append(bnd[ii][0]/scl[ii])
            x_up_bound.append(bnd[ii][1]/scl[ii])
    
        for ii in range(0,len(con)):
            edge.append(scaled_constraints[ii])
            if con[ii][1]=='<':
                con_up_edge.append(edge[ii])
                con_low_edge.append(-np.inf)
            elif con[ii][1]=='>':
                con_up_edge.append(np.inf)
                con_low_edge.append(edge[ii])
            elif con[ii][1]=='=':
                con_up_edge.append(edge[ii])
                con_low_edge.append(edge[ii])
    
        x_low_bound  = np.array(x_low_bound)
        x_up_bound   = np.array(x_up_bound)
        con_up_edge  = np.array(con_up_edge)         
        con_low_edge = np.array(con_low_edge)        
    
        return (x,scaled_constraints,x_low_bound,x_up_bound,con_up_edge,con_low_edge)    
    
    ## @ingroup Optimization-Package_Setups
    def initialize_opt_vals(self,opt_prob,obj,inp,x_low_bound,x_up_bound,con_low_edge,con_up_edge,nam,con,x_eval):
        """Sets up the optimization values 
    
        Assumptions:
        N/A
    
        Source:
        N/A
    
        Inputs:
        opt_prob         [pyopt_problem()]
        obj              [float]
        inp              [array]
        x_low_bound      [array]
        x_up_bound       [array]
        con_low_edge     [array]
        con_up_edge      [array]
        nam              [list of str]
        con              [array]
        x_eval           [array]
        
        Outputs:
        N/A
        
        Properties Used:
        N/A    
        
        """        
        
        for ii in range(len(obj)):
            opt_prob.addObj('f',100) 
        for ii in range(0,len(inp)):
            vartype = 'c'
            if x_eval is None:
                opt_prob.addVar(nam[ii],vartype,lower=x_low_bound[ii],upper=x_up_bound[ii]) 
            else:
                opt_prob.addVar(nam[ii],vartype,lower=x_low_bound[ii],upper=x_up_bound[ii],value=x_eval[ii])    
        for ii in range(0,len(con)):
            if con[ii][1]=='<':
                opt_prob.addCon(nam[ii], type='i', upper=con_up_edge[ii])
            elif con[ii][1]=='>':
                opt_prob.addCon(nam[ii], type='i', lower=con_low_edge[ii],upper=np.inf)
            elif con[ii][1]=='=':
                opt_prob.addCon(nam[ii], type='e', equal=con_up_edge[ii])        
                
        return
    
    def unpack_constraints_slsqp(self,x,con_ind,sign,edge,problem,cons_surrogate):
        
        if np.any(np.isnan(x)):
            return np.inf        
        
        const = problem.all_constraints(x).tolist()
        
        cons_addition = cons_surrogate.predict(np.atleast_2d(x))
        
        const = const + cons_addition
        const_list = const.tolist()[0]        
        
        con = (const_list[con_ind]-edge)*sign
        
        return con
    
    def initialize_opt_vals_SLSQP(self,obj,inp,x_low_bound,x_up_bound,con_low_edge,con_up_edge,nam,con,x_eval,problem,cons_surr):
        # Initialize variables according to SLSQP requirements
        x0 = x_eval
        slsqp_con_list = []
        for i,c in enumerate(con):
            c_dict = {}
            if c[1] == '<' or c[1] == '>':
                c_dict['type'] = 'ineq'
            elif c[1] == '=':
                c_dict['type'] = 'eq'
            else:
                raise ValueError('Constraint specification not recognized.')
            c_dict['fun'] = self.unpack_constraints_slsqp
            if c[1] == '<':
                c_dict['args'] = [i,-1,c[2],problem,cons_surr]
            else:
                c_dict['args'] = [i,1,c[2],problem,cons_surr]
            slsqp_con_list.append(c_dict)
        
        return x0,slsqp_con_list
    
    def initialize_opt_vals_SHGO(self,obj,inp,x_low_bound,x_up_bound,con_low_edge,con_up_edge,nam,con,problem,cons_surr):
        # Initialize variables according to SLSQP requirements
        bound_list = []
        for i in range(len(x_low_bound)):
            bound_list.append((x_low_bound[i],x_up_bound[i]))
        slsqp_con_list = []
        for i,c in enumerate(con):
            c_dict = {}
            if c[1] == '<' or c[1] == '>':
                c_dict['type'] = 'ineq'
            elif c[1] == '=':
                c_dict['type'] = 'eq'
            else:
                raise ValueError('Constraint specification not recognized.')
            c_dict['fun'] = self.unpack_constraints_slsqp
            if c[1] == '<':
                c_dict['args'] = [i,-1,c[2],problem,cons_surr]
            else:
                c_dict['args'] = [i,1,c[2],problem,cons_surr]
            slsqp_con_list.append(c_dict)
        
        return bound_list,slsqp_con_list    
    
    ## @ingroup Optimization-Package_Setups
    def run_objective_optimization(self,opt_prob,problem,f_additive_surrogate,g_additive_surrogate):
        """Runs SNOPT to optimize
    
        Assumptions:
        N/A
    
        Source:
        N/A
    
        Inputs:
        opt_prob             [pyopt_problem()]
        problem              [nexus()]
        f_additive_surrogate [fun()]  
        g_additive_surrogate [fun()]
        
        Outputs:
        fOpt                 [float]
        xOpt                 [array]
        
        Properties Used:
        N/A    
        
        """      
        
        opt = pyOpt.pySNOPT.SNOPT()
    
        problem.fidelity_level = 1
        outputs = opt(opt_prob, sens_type='FD',problem=problem, \
                      obj_surrogate=f_additive_surrogate,cons_surrogate=g_additive_surrogate)#, sens_step = sense_step)  
        fOpt = outputs[0][0]
        xOpt = outputs[1]
        
        return fOpt, xOpt


# RCAIDE/Library/Compoments/Airfoils/Airfoil.py
# 
# 
# Created:  Mar 2024, M. Clarke 

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ---------------------------------------------------------------------------------------------------------------------- 
# RCAIDE imports     
from RCAIDE.Library.Components   import Component 
 
# ---------------------------------------------------------------------------------------------------------------------- 
#  Airfoil
# ---------------------------------------------------------------------------------------------------------------------- 
 
class Airfoil(Component):
    """
    Base class for defining airfoil geometries and their aerodynamic characteristics

    Attributes
    ----------
    tag : str
        Identifier for the airfoil (default: 'Airfoil')
        
    coordinate_file : str, optional
        Path to file containing airfoil coordinates (default: None)
        
    geometry : array_like, optional
        Array containing x,y coordinates defining the airfoil shape (default: None)
        
    polar_files : list of str, optional
        Paths to files containing aerodynamic polar data (default: None)
        
    polars : dict, optional
        Dictionary containing aerodynamic coefficient data (default: None)
        
    prev : Airfoil, optional
        Reference to previous airfoil in a wing (default: None)
        
    next : Airfoil, optional
        Reference to next airfoil in a wing (default: None)
        
    number_of_points : int
        Number of points used to discretize the airfoil geometry (default: 201)

    Notes
    -----
    The Airfoil class serves as a base class for more specific airfoil implementations.
    It provides basic structure for storing geometry and aerodynamic data, which can be
    loaded from files or generated programmatically.
    
    **Definitions**
    
    'Polar Data'
        Aerodynamic coefficient data (typically cl, cd, cm) as a function of angle of attack

    See Also
    --------
    RCAIDE.Library.Components.Airfoils.NACA_4_Series_Airfoil
        Implementation for NACA 4-series airfoils
    """
    
    def __defaults__(self):
        """This sets the default values.
    
        Assumptions:
            None
        
        Source:
            None
        """         
        
        self.tag                        = 'airfoil' 
        self.coordinate_file            = None     
        self.geometry                   = None
        self.polar_files                = None
        self.polars                     = None
        self.prev                       = None
        self.next                       = None
        self.number_of_points           = 201
       


# RCAIDE/Library/Methods/Aerodynamics/Common/Drag/drag_divergence.py
# (c) Copyright 2023 Aerospace Research Community LLC
# 
# Created:  Jun 2024, M. Clarke 

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ---------------------------------------------------------------------------------------------------------------------- 
   
from RCAIDE.Library.Components.Wings          import Main_Wing 

# package imports
import numpy as np

# ----------------------------------------------------------------------------------------------------------------------
# drag_divergence
# ----------------------------------------------------------------------------------------------------------------------
def drag_divergence(Mach,geometry, Cl):
    """Use drag divergence Mach number to determine drag for subsonic speeds

    Assumptions:
    Basic fit, subsonic

    Source:
    http://aerodesign.stanford.edu/aircraftdesign/aircraftdesign.html (Stanford AA241 A/B Course Notes)
    Concorde data can be found in "Supersonic drag reduction technology in the scaled supersonic 
    experimental airplane project by JAXA" by Kenji Yoshida

    Args:
    wing.
      thickness_to_chord    [-]     
      sweeps.quarter_chord  [radians]
      high_mach             [Boolean]
      areas.reference       [m^2]

    Returns:
    cd_c                    [-]
    Machc                   [-]
    MDiv                    [-]
  
    """

    t_c_w     = 0
    sweep_w   = 0
    high_mach = False 
    for wing in  geometry.wings:
        if isinstance(wing, Main_Wing):  
            t_c_w     = wing.thickness_to_chord
            sweep_w   = wing.sweeps.quarter_chord
            high_mach =  wing.high_mach
        
    # Check if the wing is designed for high subsonic cruise
    # If so use arbitrary divergence point as correlation will not work
    if high_mach == True: 
        # Divergence Mach number, fit to Concorde data 
        Mcc    =  0.95 * np.ones_like(Mach)  
    else:  
        # Get effective Cl_wings and sweep 
        tc = t_c_w / np.cos(sweep_w)
        cl = Cl/ (np.cos(sweep_w) ** 2)

        # Compressibility drag based on regressed fits from AA241
    
        mcc_cos_ws = 0.922321524499352       \
                   - 1.153885166170620*tc    \
                   - 0.304541067183461*cl    \
                   + 0.332881324404729*tc*tc \
                   + 0.467317361111105*tc*cl \
                   + 0.087490431201549*cl*cl
         
        # Crest-critical Mach number, corrected for wing sweep
        Mcc = mcc_cos_ws/ np.cos(sweep_w)      

    # Divergence ratio
    mo_Mach = Mach/Mcc

    # Compressibility correlation, Shevell
    dcdc_cos3g = 0.0019*mo_Mach**14.641

    # Compressibility drag 
    # Sweep correlation cannot be used if the wing has a high Mach design
    if high_mach is True:
        cd_c = dcdc_cos3g
    else:
        cd_c = dcdc_cos3g * (np.cos(sweep_w)**3)
          
    return cd_c 



# RCAIDE/Framework/Analyses/Mission/Sequential_Segments.py
# 
# 
# Created:  Jul 2023, M. Clarke 

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------  
# RCAIDE imports   
import RCAIDE
from RCAIDE.Library.Mission.Common.Segments    import  sequential_segments
from RCAIDE.Library.Mission.Common.Pre_Process import  aerodynamics,stability, energy,emissions, set_residuals_and_unknowns
from RCAIDE.Framework.Core                               import Container as ContainerBase
from RCAIDE.Framework.Analyses                           import Process 
from . import Segments

# ----------------------------------------------------------------------------------------------------------------------
# ANALYSIS
# ----------------------------------------------------------------------------------------------------------------------  -Mission
class Sequential_Segments(Segments.Segment.Container):
    """ Solves each segment one at time
    
        Assumptions:
        None
        
        Source:
        None
    """
    
    def __defaults__(self):
        """This sets the default values.
    
            Assumptions:
            None
    
            Source:
            N/A
    
            Inputs:
            None
    
            Outputs:
            None
    
            Properties Used:
            None
        """          

        self.tag = 'mission'
        
        #   Initialize   
        self.process.initialize                                = Process() 
        self.process.initialize.aero                           = aerodynamics
        self.process.initialize.stability                      = stability
        self.process.initialize.energy                         = energy
        self.process.initialize.emissions                      = emissions
        self.process.initialize.set_residuals_and_unknowns     = set_residuals_and_unknowns
 
        #   Converge 
        self.process.converge    = sequential_segments
         
        #   Iterate     
        del self.process.iterate  

        return  

    def pre_process(self):
        """ This executes the entire process
    
            Assumptions:
            None
    
            Source:
            N/A
    
            Inputs:
            State  [Data()]
    
            Outputs:
            State  [Data()]
    
            Properties Used:
            None
        """   
        self.process.pre_process(self)
         
        return self 

                        
    def evaluate(self,state=None):
        """ This executes the entire process
    
            Assumptions:
            None
    
            Source:
            N/A
    
            Inputs:
            State  [Data()]
    
            Outputs:
            State  [Data()]
    
            Properties Used:
            None
        """  
        
        if state is None:
            state = self.state
        self.process(self)
        return self     
        
    
# ----------------------------------------------------------------------
#   Container Class
# ----------------------------------------------------------------------
class Container(ContainerBase):
    """ Container for mission
    
        Assumptions:
        None
        
        Source:
        None
    """    
    
    def evaluate(self,state=None):
        """ Go through the missions, run through them, save the results
    
            Assumptions:
            None
    
            Source:
            N/A
    
            Inputs:
            state   [Data()]
    
            Outputs:
            Results [Data()]
    
            Properties Used:
            None
        """         
        results = RCAIDE.Framework.Core.Data()
        
        for key,mission in self.items():
            result = mission.evaluate(state)
            results[key] = result
            
        return results
    
    def finalize(self):
        """ Stub
    
            Assumptions:
            None
    
            Source:
            N/A
    
            Inputs:
            None
    
            Outputs:
            None
    
            Properties Used:
            None
            """          
        pass

# Link container
Sequential_Segments.Container = Container


# RCAIDE/Framework/Analyses/Noise/Noise.py
# 
# 
# Created:  Jul 2023, M. Clarke

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ---------------------------------------------------------------------------------------------------------------------- 
from RCAIDE.Framework.Core import Data
from RCAIDE.Framework.Analyses import Analysis

# package imports
import numpy as np 
# ----------------------------------------------------------------------
#  Noise
# ----------------------------------------------------------------------
class Noise(Analysis):
    """ RCAIDE.Framework.Analyses.Noise.Noise()
    
        The Top Level Noise Analysis Class
        
            Assumptions:
            None
            
            Source:
            N/A
    """
    def __defaults__(self):
        """This sets the default values and methods for the analysis.
    
            Assumptions:
            None
    
            Source:
            N/A
    
            Inputs:
            None
    
            Outputs:
            None
    
            Properties Used:
            N/A
            """                   
        self.tag                                             = 'Noise'        
                                         
        self.vehicle                                         = Data()
        self.settings                                        = Data()

        self.settings                                        = self.settings
        self.settings.harmonics                              = np.arange(1,30) 
        self.settings.flyover                                = False    
        self.settings.approach                               = False
        self.settings.sideline                               = False
        self.settings.sideline_x_position                    = 0 
        self.settings.print_noise_output                     = False  
        self.settings.mean_sea_level_altitude                = True 
        self.settings.aircraft_origin_location               = np.array([0,0,0])
        self.settings.aircraft_destination_location          = np.array([0,0,0])
        self.settings.aircraft_origin_coordinates            = np.array([0.0,0.0])
        self.settings.aircraft_destination_coordinates       = np.array([0.0,0.0])
        self.settings.topography_file                        = None
        self.settings.microphone_locations                   = None   
        self.settings.microphone_coordinates                 = None
        self.settings.microphone_x_resolution                = 11 
        self.settings.microphone_y_resolution                = 11 
        self.settings.noise_times_steps                      = 101 
        self.settings.number_of_microphone_in_stencil        = 10
        self.settings.microphone_min_x                       = 0  
        self.settings.microphone_max_x                       = 1000 
        self.settings.microphone_min_y                       = -100  
        self.settings.microphone_max_y                       = 100  
        
        self.settings.noise_hemisphere                       = False 
        self.settings.noise_hemisphere_radius                = 20  
        epsilon =  1E-5
        self.settings.noise_hemisphere_phi_angles            = np.linspace(epsilon + (np.pi/2),   np.pi-epsilon,6)   
        self.settings.noise_hemisphere_theta_angles          = np.linspace(epsilon + 0        ,   2*np.pi-epsilon,12)  
                
        # settings for acoustic frequency resolution
        self.settings.center_frequencies                     = np.array([16,20,25,31.5,40, 50, 63, 80, 100, 125, 160, 200, 250, 315, 400, \
                                                                    500, 630, 800, 1000, 1250, 1600, 2000, 2500, 3150,
                                                                    4000, 5000, 6300, 8000, 10000])        
        self.settings.lower_frequencies                      = np.array([14,18,22.4,28,35.5,45,56,71,90,112,140,180,224,280,355,450,560,710,\
                                                                    900,1120,1400,1800,2240,2800,3550,4500,5600,7100,9000 ])
        self.settings.upper_frequencies                      = np.array([18,22.4,28,35.5,45,56,71,90,112,140,180,224,280,355,450,560,710,900,1120,\
                                                                 1400,1800,2240,2800,3550,4500,5600,7100,9000,11200 ])
        
        return
    
        
        
    def evaluate(self,state):
        """The default evaluate function.

        Assumptions:
        None

        Source:
        N/A

        Inputs:
        None

        Outputs:
        results   <Results class> (empty)

        Properties Used:
        N/A
        """           
        
        results = Data()
        
        return results 


# RCAIDE/Library/Methods/Weights/Correlation_Buildups/Raymer/compute_main_wing_weight.py
# 
# 
# Created:  Sep 2024, M. Clarke

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------

# RCAIDE 
from RCAIDE.Framework.Core    import Units

# python imports 
import  numpy as  np
 
# ----------------------------------------------------------------------------------------------------------------------
# Main Wing Weight 
# ----------------------------------------------------------------------------------------------------------------------
def compute_main_wing_weight(vehicle, wing):
    """ Calculate the wing weight of the aircraft based the Raymer method

    Assumptions:

    Source:
        Aircraft Design: A Conceptual Approach (2nd edition)

    Inputs:
        vehicle - data dictionary with vehicle properties                   [dimensionless]
                -.mass_properties.max_takeoff: MTOW                         [kg]
                -.flight_envelope.ultimate_load: ultimate loading factor
                -.systems.accessories: type of aircraft (short-range, commuter
                                                        medium-range, long-range,
                                                        sst, cargo)
        wing    - data dictionary with specific wing properties             [dimensionless]
                -.taper: taper ratio
                -.sweeps.quarter_chord: quarter chord sweep angle           [deg]
                -.thickness_to_chord: thickness to chord
                -.aspect_ratio: aspect ratio of wing
                -.areas.reference: wing surface area                        [m^2]

    Outputs:
        weight - weight of the wing                  [kilograms]


    Properties Used:
        N/A
    """

    # unpack inputs
    taper   = wing.taper
    sweep   = wing.sweeps.quarter_chord
    area    = wing.areas.reference
    t_c_w   = wing.thickness_to_chord

    Wdg     = vehicle.mass_properties.max_takeoff / Units.lb
    Nz      = vehicle.flight_envelope.ultimate_load
    Sw      = area / Units.ft ** 2
    A       = wing.aspect_ratio
    tc_root = t_c_w
    Scsw    = Sw * .1

    if vehicle.systems.accessories == 'sst':
        sweep = 0
    W_wing = 0.0051 * (Wdg * Nz) ** .557 * Sw ** .649 * A ** .5 * tc_root ** -.4 * (1 + taper) ** .1 * np.cos(
        sweep) ** -1. * Scsw ** .1
    weight = W_wing * Units.lb

    return weight



# RCAIDE/Compoments/Booms/Boom.py
# 
# 
# Created:  Mar 2024, M. Clarke 

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ---------------------------------------------------------------------------------------------------------------------- 
# RCAIDE imports 
from RCAIDE.Library.Components import Component 
from RCAIDE.Framework.Core     import Data, Container 
 
# ----------------------------------------------------------------------------------------------------------------------
#  BOOM
# ----------------------------------------------------------------------------------------------------------------------  
class Boom(Component):
    """
    A structural boom component model for aircraft and rotorcraft applications.

    Attributes
    ----------
    tag : str
        Identifier for the boom component. Default is 'boom'.

    origin : list
        3D coordinates of the boom origin [m]. Default is [[0.0,0.0,0.0]].

    aerodynamic_center : list
        3D coordinates of the aerodynamic center [m]. Default is [0.0,0.0,0.0].

    areas : Data
        Collection of area measurements

        - front_projected : float
            Front projected area [m²]. Default is 0.0.
        - side_projected : float
            Side projected area [m²]. Default is 0.0.
        - wetted : float
            Wetted area of the boom [m²]. Default is 0.0.

    effective_diameter : float
        Effective diameter of the boom [m]. Default is 0.0.

    width : float
        Width of the boom [m]. Default is 0.0.

    heights : Data
        Collection of height measurements

        - maximum : float
            Maximum height of the boom [m]. Default is 0.0.
        - at_quarter_length : float
            Height at 25% of boom length [m]. Default is 0.0.
        - at_three_quarters_length : float
            Height at 75% of boom length [m]. Default is 0.0.
        - at_wing_root_quarter_chord : float
            Height at wing root quarter chord [m]. Default is 0.0.
        - at_vertical_root_quarter_chord : float
            Height at vertical root quarter chord [m]. Default is 0.0.

    x_rotation : float
        Rotation angle around x-axis [rad]. Default is 0.0.

    y_rotation : float
        Rotation angle around y-axis [rad]. Default is 0.0.

    z_rotation : float
        Rotation angle around z-axis [rad]. Default is 0.0.

    lengths : Data
        Collection of length measurements

        - nose : float
            Length of the nose section [m]. Default is 0.0.
        - total : float
            Total length of the boom [m]. Default is 0.0.
        - cabin : float
            Length of the cabin section [m]. Default is 0.0.
        - fore_space : float
            Length of space in front [m]. Default is 0.0.
        - aft_space : float
            Length of space in rear [m]. Default is 0.0.

    fineness : Data
        Fineness ratios

        - nose : float
            Fineness ratio of nose. Default is 0.0.
        - tail : float
            Fineness ratio of tail. Default is 0.0.

    differential_pressure : float
        Pressure differential across the boom [Pa]. Default is 0.0.

    vsp_data : Data
        Vehicle Sketch Pad related data

        - xsec_surf_id : str
            VSP cross-section surface identifier. Default is ''.
        - xsec_num : int
            Number of cross-sections in boom geometry. Default is None.

    segments : Container
        Container for boom segments. Default is empty container.

    Notes
    -----
    The Boom class provides a comprehensive framework for modeling structural
    booms in aircraft and rotorcraft, including:

    * Geometric definition
    * Cross-sectional properties
    * Aerodynamic characteristics
    * Structural interfaces
    * VSP integration
    * Segmentation capabilities

    **Definitions**

    'Fineness Ratio'
        Ratio of length to maximum diameter
    'Wetted Area'
        Total surface area exposed to airflow
    'Differential Pressure'
        Pressure difference between inside and outside of boom
    """
    
    def __defaults__(self):
        """ :meta private:"""
        #This sets the default values.
    
        #Assumptions:
        #    None
        
        #Source:
        #    None
              
        
        self.tag                                    = 'boom'
        self.origin                                 = [[0.0,0.0,0.0]]
        self.aerodynamic_center                     = [0.0,0.0,0.0]  
                 
        self.areas                                  = Data()
        self.areas.front_projected                  = 0.0
        self.areas.side_projected                   = 0.0
        self.areas.wetted                           = 0.0
                         
        self.effective_diameter                     = 0.0
        self.width                                  = 0.0 
                         
        self.heights                                = Data()
        self.heights.maximum                        = 0.0
        self.heights.at_quarter_length              = 0.0
        self.heights.at_three_quarters_length       = 0.0
        self.heights.at_wing_root_quarter_chord     = 0.0
        self.heights.at_vertical_root_quarter_chord = 0.0
        
        self.x_rotation                             = 0.0
        self.y_rotation                             = 0.0
        self.z_rotation                             = 0.0
             
        self.lengths                                = Data()
        self.lengths.nose                           = 0.0 
        self.lengths.total                          = 0.0
        self.lengths.cabin                          = 0.0
        self.lengths.fore_space                     = 0.0
        self.lengths.aft_space                      = 0.0
                 
        self.fineness                               = Data()
        self.fineness.nose                          = 0.0
        self.fineness.tail                          = 0.0
             
        self.differential_pressure                  = 0.0 
              
        # For VSP     
        self.vsp_data                               = Data()
        self.vsp_data.xsec_surf_id                  = ''    # There is only one XSecSurf in each VSP geom.
        self.vsp_data.xsec_num                      = None  # Number if XSecs in rotor_boom geom.
                        
        self.segments                               = Container()
        
    def append_segment(self,segment):
        """
        Assumptions:
           None
            
        Source:
           None
        
        Args:
           self       : boom                  [unitless]
           segment    : cross-section of boom [unitless]   
            
        Outputs:
           None 
        """

        # Assert database type
        if not isinstance(segment,Data):
            raise Exception('input component must be of type Data()')

        # Store data
        self.segments.append(segment) 
        
        return 

class Container(Component.Container):
    def get_children(self):
        """ :meta private: """
        #Returns the components that can go inside
    
        #Assumptions:
        #    None
            
        #Source:
        #    None

        #Args:
        #    self       : container of booms [unitless]    
            
        #Outputs:
        #    Boom       : boom               [unitless] 
        
        return [Boom]

# ------------------------------------------------------------
#  Handle Linking
# ------------------------------------------------------------ 
Boom.Container = Container



## @ingroup Methods-Aerodynamics-Common-Fidelity_Zero-Helper_Functions
# wave_drag_lift.py
# 
# Created:  Jun 2014, T. Macdonald
# Modified: Jul 2014, T. Macdonald
#           Jan 2016, E. Botero

# ----------------------------------------------------------------------
#  Imports
# ----------------------------------------------------------------------

import numpy as np
from RCAIDE.Framework.Core import Data

# ----------------------------------------------------------------------
#   Wave Drag Lift
# ----------------------------------------------------------------------

## @ingroup Methods-Aerodynamics-Common-Fidelity_Zero-Helper_Functions
def wave_drag_lift(conditions,configuration,wing):
    """Computes wave drag due to lift

    Assumptions:
    Simplified equations

    Source:
    http://adg.stanford.edu/aa241/drag/ssdragcalc.html

    Inputs:
    conditions.freestream.mach_number        [Unitless]
    conditions.aerodynamics.lift_coefficient [Unitless]
    wing.total_length                        [m]
    wing.areas.reference                     [m^2]

    Outputs:
    wave_drag_lift                           [Unitless]

    Properties Used:
    N/A
    """  

    # Unpack
    freestream = conditions.freestream
    total_length = wing.total_length
    Sref = wing.areas.reference
    
    # Conditions
    Mc  = freestream.mach_number * 1.0

    # Length-wise aspect ratio
    ARL = total_length**2/Sref
    
    # Lift coefficient
    if wing.vertical:
        CL = np.zeros_like(conditions.aerodynamics.lift_coefficient)
    else:
        # get wing specific CL
        CL = conditions.aerodynamics.lift_breakdown.inviscid_wings_lift[wing.tag]
    
    # Computations
    x = np.pi*ARL/4
    beta = np.array([[0.0]] * len(Mc))
    beta[Mc >= 1.05] = np.sqrt(Mc[Mc >= 1.05]**2-1)
    wave_drag_lift = np.array([[0.0]] * len(Mc))
    wave_drag_lift[Mc >= 1.05] = CL[Mc >= 1.05]**2*x/4*(np.sqrt(1+(beta[Mc >= 1.05]/x)**2)-1)
    wave_drag_lift[0:len(Mc[Mc >= 1.05]),0] = wave_drag_lift[Mc >= 1.05]
    
    # Dump data to conditions
    wave_lift_result = Data(
        reference_area             = Sref   , 
        wave_drag_lift_coefficient = wave_drag_lift ,
        length_AR                  = ARL,
    )

    return wave_drag_lift



# RCAIDE/Library/Missions/Common/helper_functions.py
# 
# 
# Created:  Jul 2023, M. Clarke
 
# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------  
# RCAIDE imports 
import RCAIDE 
from RCAIDE.Framework.Core  import Data 

def pre_process(mission): 
    for tag,segment in mission.segments.items():     
        segment.pre_process()

def sequential_segments(mission):  
    
    last_tag = None
    for tag,segment in mission.segments.items(): 
        if last_tag:
            segment.state.initials = mission.segments[last_tag].state
        last_tag = tag        
        
        segment.process.initialize.expand_state(segment) 
        segment.process.initialize.expand_state = RCAIDE.Library.Methods.skip        
        segment.evaluate()
        
def update_segments(mission):   
    for tag,segment in mission.segments.items():
        segment.post_process() 
        
def merge_segment_states(mission): 
    mission.state.update(mission.merged())
    
def unpack_segments(mission): 
    # Build a dict with the sections, sections start at 0
    counter = Data()
    
    for key in mission.state.unknowns.keys():
        counter[key] = 0

    for i, segment in enumerate(mission.segments):
        for key in segment.state.unknowns.keys():
            if key=='tag':
                continue
            points = segment.state.unknowns[key].size
            segment.state.unknowns[key] = mission.state.unknowns[key][counter[key]:counter[key]+points]
            counter[key] = counter[key]+points
            
    return
            
            


# RCAIDE/Library/Methods/Geometry/Airfoil/import_airfoil_geometry.py

# ----------------------------------------------------------------------
#  Imports
# ---------------------------------------------------------------------- 
from RCAIDE.Framework.Core import  Data 
import numpy as np
from scipy import interpolate

# ----------------------------------------------------------------------------------------------------------------------
# import_airfoil_geometry
# ----------------------------------------------------------------------------------------------------------------------   
def import_airfoil_geometry(airfoil_geometry_file, npoints = 201,surface_interpolation = 'cubic'):
    """This imports an airfoil geometry from a text file  and store
    the coordinates of upper and lower surfaces as well as the mean
    camberline
    
    Assumptions:
    Works for Selig and Lednicer airfoil formats. Automatically detects which format based off first line of data. Assumes it is one of those two.
    Source:
    airfoiltools.com/airfoil/index - method for determining format and basic error checking
    Inputs:
    airfoil_geometry_files   <list of strings>
    surface_interpolation   - type of interpolation used in the SciPy function. Preferable options are linear, quardratic and cubic. 
    Full list of options can be found here : 
    https://docs.scipy.org/doc/scipy/reference/generated/scipy.interpolate.interp1d.html#scipy.interpolate.interp1d
    Outputs:
    airfoil_data.
        thickness_to_chord 
        x_coordinates 
        y_coordinates
        x_upper_surface
        x_lower_surface
        y_upper_surface
        y_lower_surface
        camber_coordinates  
    Properties Used:
    N/A
    """

    if npoints%2 != 1:
        npoints+= 1
        print('Number of points must be odd, changing to ' + str(npoints) + ' points')      
    
    geometry     = Data()
    half_npoints = npoints//2         
 
    # Open file and read column names and data block
    f = open(airfoil_geometry_file) 

    # Extract data
    data_block = f.readlines()
    try:
        # Check for header block
        first_element = float(data_block[0][0])
        if first_element == 1.:
            lednicer_format = False
    except:
        # Check for format line and remove header block
        format_line = data_block[1]

        # Check if it's a Selig or Lednicer file
        try:
            format_flag = float(format_line.strip().split()[0])
        except:
            format_flag = float(format_line.strip().split(',')[0])
            
        if format_flag > 1.01: # Amount of wiggle room per airfoil tools
            lednicer_format = True
            # Remove header block
            data_block      = data_block[3:]
        else:
            lednicer_format = False   
            # Remove header block
            data_block = data_block[1:]
            
    # Close the file
    f.close() 

    if lednicer_format:
        x_up_surf = []
        y_up_surf = []
        x_lo_surf = []
        y_lo_surf = []
        
        # Loop through each value: append to each column
        upper_surface_flag = True
        for line_count , line in enumerate(data_block): 
            #check for blank line which signifies the upper/lower surface division 
            line_check = data_block[line_count].strip()
            if line_check == '':
                upper_surface_flag = False
                continue
            if upper_surface_flag:
                x_up_surf.append(float(data_block[line_count].strip().split()[0])) 
                y_up_surf.append(float(data_block[line_count].strip().split()[1])) 
            else:                              
                x_lo_surf.append(float(data_block[line_count].strip().split()[0])) 
                y_lo_surf.append(float(data_block[line_count].strip().split()[1]))   

    else:
        x_up_surf_rev  = []
        y_up_surf_rev  = []
        x_lo_surf      = []
        y_lo_surf      = []
        
        # Loop through each value: append to each column
        upper_surface_flag = True
        for line_count , line in enumerate(data_block): 
            #check for line which starts with 0., which should be the split between upper and lower in selig
            line_check = data_block[line_count].strip()
            
            # Remove any commas
            line_check = line_check.replace(',','')
            
            if float(line_check.split()[0]) == 0.:
                x_up_surf_rev.append(float(data_block[line_count].strip().replace(',','').split()[0])) 
                y_up_surf_rev.append(float(data_block[line_count].strip().replace(',','').split()[1]))

                x_lo_surf.append(float(data_block[line_count].strip().replace(',','').split()[0])) 
                y_lo_surf.append(float(data_block[line_count].strip().replace(',','').split()[1])) 

                upper_surface_flag = False
                continue

            if upper_surface_flag:
                x_up_surf_rev.append(float(data_block[line_count].strip().replace(',','').split()[0])) 
                y_up_surf_rev.append(float(data_block[line_count].strip().replace(',','').split()[1])) 
            else:                              
                x_lo_surf.append(float(data_block[line_count].strip().replace(',','').split()[0])) 
                y_lo_surf.append(float(data_block[line_count].strip().replace(',','').split()[1]))
            
            
            if upper_surface_flag ==True:
                # check if next line flips without x-coordinate going to 0
                next_line  = data_block[line_count+1].strip()
                next_line  = next_line.replace(',','')
            
                if next_line.split()[0]>line_check.split()[0] and float(next_line.split()[0]) >0.:
                    upper_surface_flag = False
                
        # Upper surface values in Selig format are reversed from Lednicer format, so fix that
        x_up_surf_rev.reverse()
        y_up_surf_rev.reverse()

        x_up_surf = x_up_surf_rev
        y_up_surf = y_up_surf_rev 

    x_up_surf = np.array(x_up_surf)
    x_lo_surf = np.array(x_lo_surf)
    y_up_surf = np.array(y_up_surf)
    y_lo_surf = np.array(y_lo_surf)  
    
    # Check for extra zeros (OpenVSP exports extra zeros)
    if len(np.unique(x_up_surf))!=len(x_up_surf):
        x_up_surf = x_up_surf[1:]
        x_lo_surf = x_lo_surf[1:]
        y_up_surf = y_up_surf[1:]
        y_lo_surf = y_lo_surf[1:]

    # create custom spacing for more points and leading and trailing edge
    t            = np.linspace(0,4,npoints-1)
    delta        = 0.25 
    A            = 5
    f            = 0.25
    smoothsq     = 5 + (2*A/np.pi) *np.arctan(np.sin(2*np.pi*t*f + np.pi/2)/delta) 
    dim_spacing  = np.append(0,np.cumsum(smoothsq)/sum(smoothsq))
    
    # compute thickness, camber and concatenate coodinates 
    x_data        = np.hstack((x_lo_surf[::-1], x_up_surf[1:])) 
    y_data        = np.hstack((y_lo_surf[::-1], y_up_surf[1:]))   
    tck,u         = interpolate.splprep([x_data,y_data],k=3,s=0) 
    out           = interpolate.splev(dim_spacing,tck) 
    x_data        = out[0]   
    y_data        = out[1]  
    
    # shift points to leading edge (x = 0, y = 0)
    x_delta  = min(x_data)
    x_data   = x_data - x_delta 
    
    arg_min  = np.argmin(x_data) 
    y_delta  = y_data[arg_min]
    y_data   = y_data - y_delta
    
    if (x_data[arg_min] == 0) and (y_data[arg_min]  == 0): 
        x_data[arg_min]  = 0  
        y_data[arg_min]  = 0 
    
    # make sure points start and end at x = 1.0
    x_data[0]  = 1.0
    x_data[-1] = 1.0
    
    # make sure a small gap at trailing edge
    if (y_data[0] == y_data[-1]): 
        y_data[0]          = y_data[0]  - 1E-4
        y_data[-1]         = y_data[-1] + 1E-4
        
    # thicknes and camber distributions require equal points     
    x_up_surf_old  = np.array(x_up_surf)   
    arrx_up_interp = interpolate.interp1d(np.arange(x_up_surf_old.size),x_up_surf_old, kind=surface_interpolation)
    x_up_surf_new  = arrx_up_interp(np.linspace(0,x_up_surf_old.size-1,half_npoints))    
 
    x_lo_surf_old  = np.array(x_lo_surf) 
    arrx_lo_interp = interpolate.interp1d(np.arange(x_lo_surf_old.size),x_lo_surf_old, kind=surface_interpolation )
    x_lo_surf_new  = arrx_lo_interp(np.linspace(0,x_lo_surf_old.size-1,half_npoints)) 
 
    # y coordinate s 
    y_up_surf_old  = np.array(y_up_surf)   
    arry_up_interp = interpolate.interp1d(np.arange(y_up_surf_old.size),y_up_surf_old, kind=surface_interpolation)
    y_up_surf_new  = arry_up_interp(np.linspace(0,y_up_surf_old.size-1,half_npoints))    
 
    y_lo_surf_old  = np.array(y_lo_surf) 
    arry_lo_interp = interpolate.interp1d(np.arange(y_lo_surf_old.size),y_lo_surf_old, kind=surface_interpolation)
    y_lo_surf_new  = arry_lo_interp(np.linspace(0,y_lo_surf_old.size-1,half_npoints)) 

    # compute thickness, camber and concatenate coodinates 
    thickness      = y_up_surf_new - y_lo_surf_new
    camber         = y_lo_surf_new + thickness/2  
    max_t          = np.max(thickness)
    max_c          = max(x_data) - min(x_data)
    t_c            = max_t/max_c 
    
    geometry.thickness_to_chord = t_c
    geometry.max_thickness      = max_t     
    geometry.x_coordinates      = x_data   
    geometry.y_coordinates      = y_data      
    geometry.x_upper_surface    = x_up_surf_new 
    geometry.x_lower_surface    = x_lo_surf_new 
    geometry.y_upper_surface    = y_up_surf_new 
    geometry.y_lower_surface    = y_lo_surf_new              
    geometry.camber_coordinates = camber
         
    return geometry


## @ingroup Optimization-Package_Setups
# pyopt_setup.py
#
# Created:  Jul 2015, E. Botero
# Modified: Feb 2016, M. Vegh
#           May 2021, E. Botero 

# ----------------------------------------------------------------------
#  Imports
# ----------------------------------------------------------------------

# RCAIDE imports
import numpy as np
from RCAIDE.Framework.Optimization.Common import helper_functions as help_fun


# ----------------------------------------------------------------------
#  Pyopt_Solve
# ----------------------------------------------------------------------

## @ingroup Optimization-Package_Setups
def Pyopt_Solve(problem,solver='SNOPT',FD='single', sense_step=1.0E-6,  nonderivative_line_search=False):
    """ This converts your RCAIDE Nexus problem into a PyOpt optimization problem and solves it
        PyOpt has many algorithms, they can be switched out by using the solver input. 

        Assumptions:
        None

        Source:
        N/A

        Inputs:
        problem                   [nexus()]
        solver                    [str]
        FD (parallel or single)   [str]
        sense_step                [float]
        nonderivative_line_search [bool]

        Outputs:
        outputs                   [list]

        Properties Used:
        None
    """      
   
    # Have the optimizer call the wrapper
    mywrap = lambda x:PyOpt_Problem(problem,x)
   
    inp = problem.optimization_problem.inputs
    obj = problem.optimization_problem.objective
    con = problem.optimization_problem.constraints
   
    if FD == 'parallel':
        from mpi4py import MPI
        comm = MPI.COMM_WORLD
        myrank = comm.Get_rank()      
   
    # Instantiate the problem and set objective
    import pyOpt
    opt_prob = pyOpt.Optimization('RCAIDE',mywrap)
    for ii in range(len(obj)):
        opt_prob.addObj(obj[ii,0])    
       
    # Set inputs
    nam  = inp[:,0] # Names
    ini  = inp[:,1] # Initials
    bndl = inp[:,2] # Bounds
    bndu = inp[:,3] # Bounds
    scl  = inp[:,4] # Scale
    typ  = inp[:,5] # Type
   
    # Pull out the constraints and scale them
    bnd_constraints = help_fun.scale_const_bnds(con)
    scaled_constraints = help_fun.scale_const_values(con,bnd_constraints)
    x   = ini/scl
   
    for ii in range(0,len(inp)):
        lbd = (bndl[ii]/scl[ii])
        ubd = (bndu[ii]/scl[ii])
        #if typ[ii] == 'continuous':
        vartype = 'c'
        #if typ[ii] == 'integer':
            #vartype = 'i'
        opt_prob.addVar(nam[ii],vartype,lower=lbd,upper=ubd,value=x[ii])
       
    # Setup constraints  
    for ii in range(0,len(con)):
        name = con[ii][0]
        edge = scaled_constraints[ii]
       
        if con[ii][1]=='<':
            opt_prob.addCon(name, type='i', upper=edge)
        elif con[ii][1]=='>':
            opt_prob.addCon(name, type='i', lower=edge,upper=np.inf)
        elif con[ii][1]=='=':
            opt_prob.addCon(name, type='e', equal=edge)

    # Finalize problem statement and run  
    print(opt_prob)
   
    if solver == 'SNOPT':
        import pyOpt.pySNOPT
        opt = pyOpt.pySNOPT.SNOPT()
        CD_step = (sense_step**2.)**(1./3.)  #based on SNOPT Manual Recommendations
        opt.setOption('Function precision', sense_step**2.)
        opt.setOption('Difference interval', sense_step)
        opt.setOption('Central difference interval', CD_step)

    elif solver == 'COBYLA':
        import pyOpt.pyCOBYLA
        opt = pyOpt.pyCOBYLA.COBYLA() 
        
    elif solver == 'SLSQP':
        import pyOpt.pySLSQP
        opt = pyOpt.pySLSQP.SLSQP()
        opt.setOption('MAXIT', 200)
    elif solver == 'KSOPT':
        import pyOpt.pyKSOPT
        opt = pyOpt.pyKSOPT.KSOPT()
    elif solver == 'ALHSO':
        import pyOpt.pyALHSO
        opt = pyOpt.pyALHSO.ALHSO()   
    elif solver == 'FSQP':
        import pyOpt.pyFSQP
        opt = pyOpt.pyFSQP.FSQP()
    elif solver == 'PSQP':
        import pyOpt.pyPSQP
        opt = pyOpt.pyPSQP.PSQP()    
    elif solver == 'NLPQL':
        import pyOpt.pyNLPQL
        opt = pyOpt.pyNLPQL.NLPQL()    
    elif solver == 'NSGA2':
        import pyOpt.pyNSGA2
        opt = pyOpt.pyNSGA2.NSGA2(pll_type='POA') 
    elif solver == 'MIDACO':
        import pyOpt.pyMIDACO
        opt = pyOpt.pyMIDACO.MIDACO(pll_type='POA')     
    elif solver == 'ALPSO':
        import pyOpt.pyALPSO
        #opt = pyOpt.pyALPSO.ALPSO(pll_type='DPM') #this requires DPM, which is a parallel implementation
        opt = pyOpt.pyALPSO.ALPSO()
    if nonderivative_line_search==True:
        opt.setOption('Nonderivative linesearch')
    if FD == 'parallel':
        outputs = opt(opt_prob, sens_type='FD',sens_mode='pgc')
        
    elif solver == 'SNOPT' or solver == 'SLSQP':
        outputs = opt(opt_prob, sens_type='FD', sens_step = sense_step)
  
    else:
        outputs = opt(opt_prob)        
   
    return outputs


# ----------------------------------------------------------------------
#  Problem Wrapper
# ----------------------------------------------------------------------

## @ingroup Optimization-Package_Setups
def PyOpt_Problem(problem,x):
    """ This wrapper runs the RCAIDE problem and is called by the PyOpt solver.
        Prints the inputs (x) as well as the objective values and constraints.
        If any values produce NaN then a fail flag is thrown.

        Assumptions:
        None

        Source:
        N/A

        Inputs:
        problem   [nexus()]
        x         [array]

        Outputs:
        obj       [float]
        cons      [array]
        fail      [bool]

        Properties Used:
        None
    """      
   
    obj   = problem.objective(x)
    const = problem.all_constraints(x).tolist()
    fail  = np.array(np.isnan(obj.tolist()) or np.isnan(np.array(const).any())).astype(int)

       
    print('Inputs')
    print(x)
    print('Obj')
    print(obj)
    print('Con')
    print(const)
   
    return obj,const,fail



# RCAIDE/Framework/Mission/Segment/Segment.py
# 
# 
# Created:  Jul 2023, M. Clarke
 
# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ---------------------------------------------------------------------------------------------------------------------- 
# RCAIDE Imports
from RCAIDE.Framework.Core import Data
from RCAIDE.Framework.Analyses                    import Analysis, Settings, Process   
from RCAIDE.Framework.Mission.Common     import State 

# ----------------------------------------------------------------------------------------------------------------------
#  ANALYSES
# ----------------------------------------------------------------------------------------------------------------------  
class Segment(Analysis):
    """ 
    """    
    
    def __defaults__(self):
        """This sets the default values.
    
            Assumptions:
            None
    
            Source:
            N/A
    
            Inputs:
            None
    
            Outputs:
            None
    
            Properties Used:
            None
        """          
        
        self.settings                      = Settings() 
        self.state                         = State() 
        self.analyses                      = Analysis.Container() 
        self.process                       = Process() 
        self.process.initialize            = Process()
          
        self.process.converge              = Process()
        self.process.iterate               = Process()
        self.process.iterate.unknowns      = Process()
        self.process.iterate.initials      = Process()
        self.process.iterate.conditions    = Process()
        self.process.iterate.residuals     = Process()
        self.process.post_process          = Process()  
        
        self.conditions = self.state.conditions 
        
        return
    
    def initialize(self):
        """ This executes the initialize process
    
            Assumptions:
            None
    
            Source:
            N/A
    
            Inputs:
            State  [Data()]
    
            Outputs:
            None
    
            Properties Used:
            None
        """        
        self.process.initialize(self)
        return
    
    def converge(self,state):
        """ This executes the converge process
    
            Assumptions:
            None
    
            Source:
            N/A
    
            Inputs:
            State  [Data()]
    
            Outputs:
            None
    
            Properties Used:
            None
        """             
        self.process.converge(self,state)    
    
    def iterate(self):
        """ This executes the iterate process
    
            Assumptions:
            None
    
            Source:
            N/A
    
            Inputs:
            State  [Data()]
    
            Outputs:
            None
    
            Properties Used:
            None
        """        
        self.process.iterate(self)
        return
    
    def post_process(self):
        """ This executes post processing
    
            Assumptions:
            None
    
            Source:
            N/A
    
            Inputs:
            State  [Data()]
    
            Outputs:
            None
    
            Properties Used:
            None
        """         
        self.process.post_process(self)
        return
    
    def evaluate(self,state=None):
        """ This executes the entire process
    
            Assumptions:
            None
    
            Source:
            N/A
    
            Inputs:
            State  [Data()]
    
            Outputs:
            State  [Data()]
    
            Properties Used:
            None
        """          
        if state is None:
            state = self.state
        self.process(self)
        return self
    
    def flight_dynamics_and_controls(self): 
        self.flight_dynamics                                             = Data()
        self.flight_dynamics.force_x                                     = False 
        self.flight_dynamics.force_y                                     = False 
        self.flight_dynamics.force_z                                     = False 
        self.flight_dynamics.moment_x                                    = False 
        self.flight_dynamics.moment_y                                    = False 
        self.flight_dynamics.moment_z                                    = False    
        
        self.assigned_control_variables                                              = Data() 
        
        self.assigned_control_variables.body_angle                                   = Data()
        self.assigned_control_variables.body_angle.active                            = False               
        self.assigned_control_variables.body_angle.initial_guess_values              = None

        self.assigned_control_variables.bank_angle                                   = Data()
        self.assigned_control_variables.bank_angle.active                            = False 
        self.assigned_control_variables.bank_angle.initial_guess_values              = None

        self.assigned_control_variables.wind_angle                                   = Data()
        self.assigned_control_variables.wind_angle.active                            = False             
        self.assigned_control_variables.wind_angle.initial_guess_values              = None   

        self.assigned_control_variables.elapsed_time                                 = Data()
        self.assigned_control_variables.elapsed_time.active                          = False                 
        self.assigned_control_variables.elapsed_time.initial_guess_values            = None  
    
        self.assigned_control_variables.velocity                                     = Data()
        self.assigned_control_variables.velocity.active                              = False                  
        self.assigned_control_variables.velocity.initial_guess_values                = None
        
        self.assigned_control_variables.acceleration                                 = Data()
        self.assigned_control_variables.acceleration.active                          = False                 
        self.assigned_control_variables.acceleration.initial_guess_values            = None
        
        self.assigned_control_variables.altitude                                     = Data()
        self.assigned_control_variables.altitude.active                              = False                 
        self.assigned_control_variables.altitude.initial_guess_values                = None
    
        self.assigned_control_variables.throttle                                     = Data() 
        self.assigned_control_variables.throttle.active                              = False                
        self.assigned_control_variables.throttle.assigned_propulsors                 = None       
        self.assigned_control_variables.throttle.initial_guess_values                = None
    
        self.assigned_control_variables.elevator_deflection                          = Data() 
        self.assigned_control_variables.elevator_deflection.active                   = False      
        self.assigned_control_variables.elevator_deflection.assigned_surfaces        = None 
        self.assigned_control_variables.elevator_deflection.initial_guess_values     = None

        self.assigned_control_variables.rudder_deflection                            = Data()
        self.assigned_control_variables.rudder_deflection.active                     = False
        self.assigned_control_variables.rudder_deflection.assigned_surfaces          = None 
        self.assigned_control_variables.rudder_deflection.initial_guess_values       = None

        self.assigned_control_variables.flap_deflection                              = Data() 
        self.assigned_control_variables.flap_deflection.active                       = False          
        self.assigned_control_variables.flap_deflection.assigned_surfaces            = None 
        self.assigned_control_variables.flap_deflection.initial_guess_values         = None

        self.assigned_control_variables.slat_deflection                              = Data() 
        self.assigned_control_variables.slat_deflection.active                       = False          
        self.assigned_control_variables.slat_deflection.assigned_surfaces            = None 
        self.assigned_control_variables.slat_deflection.initial_guess_values         = None            
    
        self.assigned_control_variables.aileron_deflection                           = Data() 
        self.assigned_control_variables.aileron_deflection.active                    = False      
        self.assigned_control_variables.aileron_deflection.assigned_surfaces         = None 
        self.assigned_control_variables.aileron_deflection.initial_guess_values      = None
    
        self.assigned_control_variables.thrust_vector_angle                          = Data() 
        self.assigned_control_variables.thrust_vector_angle.active                   = False        
        self.assigned_control_variables.thrust_vector_angle.assigned_propulsors      = None 
        self.assigned_control_variables.thrust_vector_angle.initial_guess_values     = None
        
        return     
           
# ----------------------------------------------------------------------
#  Container
# ----------------------------------------------------------------------

class Container(Segment):
    """ A container for the segment
    
        Assumptions:
        None
        
        Source:
        None
    """    
    
    def __defaults__(self):
        """This sets the default values.
    
            Assumptions:
            None
    
            Source:
            N/A
    
            Inputs:
            None
    
            Outputs:
            None
    
            Properties Used:
            None
        """          
                
        self.segments = Process()
        
        self.state = State.Container()
        
    def append_segment(self,segment):
        """ Add a SubSegment
    
            Assumptions:
            None
    
            Source:
            N/A
    
            Inputs:
            segment  [Segment()]
    
            Outputs:
            None
    
            Properties Used:
            None
        """          
        self.segments.append(segment)
        return    
        
Segment.Container = Container


# RCAIDE/Library/Methods/Weights/Correlation_Buildups/Propulsion/integrated_propulsion_general_aviation.py
# 
# 
# Created:  Sep 2024, M. Clarke 
# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------

# RCAIDE 
from RCAIDE.Framework.Core import  Units

# ----------------------------------------------------------------------------------------------------------------------
#  Integrated Propulsion Weight 
# ----------------------------------------------------------------------------------------------------------------------
def integrated_propulsion_general_aviation(piston_engine_weight,num_eng, engine_W_factor = 2.575, engine_W_exp = .922):
    """ 
        Calculate the weight of the entire propulsion system        

        Source:
                Source: Raymer, Aircraft Design, a Conceptual Approach        
                
        Inputs:
                piston_engine_weight - dry weight of a single engine                                     [kilograms]
                num_eng - total number of engines on the aircraft                                 [dimensionless]
                engine_W_factor - weight increase factor for entire integrated propulsion system [dimensionless]
        
        Outputs:
                weight - weight of the full propulsion system [kilograms]

    """     
    engine_dry = piston_engine_weight/Units.lbs
    weight     = engine_W_factor * (engine_dry**engine_W_exp)*num_eng
    mass       = weight*Units.lbs #convert to kg

    return mass


# RCAIDE/Library/Components/Fuselages/__init__.py

"""
Module containing fuselage segments for aircraft design and analysis.

See Also
--------
RCAIDE.Library.Components.Airfoils
    Related module for airfoil definitions used in aerodynamic surfaces
RCAIDE.Library.Components.Booms
    Related module for boom components that may connect to fuselage sections
"""

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------

from .Segment                    import Segment
from .Circle_Segment             import Circle_Segment
from .Ellipse_Segment            import Ellipse_Segment
from .Super_Ellipse_Segment      import Super_Ellipse_Segment
from .Rounded_Rectangle_Segment  import Rounded_Rectangle_Segment


# RCAIDE/Library/Methods/Weights/Correlation_Buildups/FLOPS/compute_fuselage_weight.py
# 
# 
# Created:  Sep 2024, M. Clarke

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------

# RCAIDE
import RCAIDE 
from RCAIDE.Framework.Core    import Units

# python imports 
import  numpy as  np
 
# ----------------------------------------------------------------------------------------------------------------------
# Fuselage Weight 
# ----------------------------------------------------------------------------------------------------------------------
def compute_fuselage_weight(vehicle):
    """ Calculate the weight of the fuselage of a transport aircraft

        Assumptions:
            NFUSE = 1, only one fuselage (it's possible to modify this in future work)
            delta_isa = 0, for pressure calculations
            Fuselage is tagged as 'fuselage'

        Source:
            The Flight Optimization System Weight Estimation Method

        Inputs:
            vehicle - data dictionary with vehicle properties                    [dimensionless]
                -.networks: data dictionary containing all propulsion properties
                -.fuselages['fuselage'].lengths.total: fuselage total length      [meters]
                -.fuselages['fuselage'].width: fuselage width                    [meters]
                -.fuselages['fuselage'].heights.maximum: fuselage maximum height [meters]
                -.flight_envelope.ultimate_load: ultimate load factor (default: 3.75)
                -.systems.accessories: type of aircraft (short-range, commuter
                                                        medium-range, long-range,
                                                        sst, cargo)
                -.mass_properties.max_takeoff: MTOW                              [kilograms]
                -.design_mach_number: design mach number for cruise flight

        Outputs:
            WFUSE - weight of the fuselage                                      [kilograms]

        Properties Used:
            N/A
    """
    
    L =  0
    for fuselage in vehicle.fuselages:
        if L < fuselage.lengths.total: 
            total_length = fuselage.lengths.total
            width        = fuselage.width
            max_height   = fuselage.heights.maximum
    
    XL  = total_length / Units.ft  # Fuselage length, ft
    DAV = (width + max_height) / 2. * 1 / Units.ft
    if vehicle.systems.accessories == "short-range" or vehicle.systems.accessories == "commuter":
        SWFUS           = np.pi * (XL / DAV - 1.7) * DAV ** 2  # Fuselage wetted area, ft**2
        ULF             = vehicle.flight_envelope.ultimate_load  # Ultimate load factor
        atmosphere      = RCAIDE.Framework.Analyses.Atmospheric.US_Standard_1976()
        atmo_data       = atmosphere.compute_values(vehicle.flight_envelope.design_cruise_altitude, 0)
        atmo_data_floor = atmosphere.compute_values(0, 0)
        DELTA           = atmo_data.pressure/atmo_data_floor.pressure
        QCRUS           = 1481.35 * DELTA * vehicle.flight_envelope.design_mach_number**2  # Cruise dynamic pressure, psf
        DG              = vehicle.mass_properties.max_takeoff / Units.lbs  # Design gross weight in lb
        WFUSE           = 0.052 * SWFUS ** 1.086 * (ULF * DG) ** 0.177 * QCRUS ** 0.241
    else: 
        NENG = 0
        FNEW = 0
        FNEF = 0 
        for network in  vehicle.networks:
            for propulsor in network.propulsors:
                if isinstance(propulsor, RCAIDE.Library.Components.Propulsors.Turbofan) or  isinstance(propulsor, RCAIDE.Library.Components.Propulsors.Turbojet):
                    NENG += 1 
                    FNEF += 1
                    if propulsor.wing_mounted: 
                        FNEW += 1  
        if vehicle.systems.accessories == 'cargo':
            CARGF = 1
        else:
            CARGF = 0  # Cargo aircraft floor factor [0 for passenger transport, 1 for cargo transport
        NFUSE = 1  # Number of fuselages
        WFUSE = 1.35 * (XL * DAV) ** 1.28 * (1 + 0.05 * FNEF) * (1 + 0.38 * CARGF) * NFUSE
    return WFUSE * Units.lbs



# RCAIDE/Library/Missions/Common/Update/energy.py
# 
# 
# Created:  Jul 2023, M. Clarke 

# ----------------------------------------------------------------------------------------------------------------------
#  Update Battery Age
# ---------------------------------------------------------------------------------------------------------------------- 
def energy(segment):  
    """Updates battery age based on operating conditions, cell temperature and time of operation.
       Source: 
       Cell specific. See individual battery cell for more details

       Assumptions:
       Cell specific. See individual battery cell for more details

       Inputs: 
       segment.
           conditions                         - conditions of battery at each segment  [unitless]
           increment_battery_age_by_one_day   - flag to increment battery cycle day    [boolean]

       Outputs:
       N/A  

       Properties Used:
       N/A 
    """  
    # loop throuh networks in vehicle 
    for network in segment.analyses.energy.vehicle.networks: 
        if 'busses' in network: 
            busses  = network.busses
            for bus in busses:
                for battery in bus.battery_modules: 
                    increment_day = segment.increment_battery_age_by_one_day
                    battery_conditions  = segment.conditions.energy[bus.tag].battery_modules[battery.tag] 
                    battery.update_battery_age(segment,battery_conditions,increment_battery_age_by_one_day = increment_day) 


# RCAIDE/Methods/Noise/Common/compute_relative_noise_evaluation_locations.py
# 
# 
# Created:  Jul 2023, M. Clarke  

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ---------------------------------------------------------------------------------------------------------------------- 

# Python package imports   
import numpy as np  
    
# ----------------------------------------------------------------------------------------------------------------------  
#  Relative Noise Evaluatation Locations
# ----------------------------------------------------------------------------------------------------------------------      
def compute_relative_noise_evaluation_locations(settings,microphone_locations, segment):
    """This computes the relative locations on the surface in the computational domain where the 
    propogated sound is computed. Vectors point from observer/microphone to aircraft/source  
            
    Assumptions: 
        Acoustic scattering is not modeled

    Source:
        N/A  

    Inputs:  
        settings.microphone_locations                - array of microphone locations on the ground  [meters] 
        segment.conditions.frames.inertial.position_vector  - position of aircraft                         [boolean]                                          

    Outputs: 
    GM_THETA   - angle measured from ground microphone in the x-z plane from microphone to aircraft 
    GM_PHI     - angle measured from ground microphone in the y-z plane from microphone to aircraft 
    RML        - relative microphone locations  
    num_gm_mic - number of ground microphones
 
    Properties Used:
        N/A       
    """       
  
    MSL_altitude      = settings.mean_sea_level_altitude
    N                 = settings.noise_times_steps
    pos               = segment.state.conditions.frames.inertial.position_vector
    
    # rediscretize time and aircraft position to get finer resolution 
    time              = segment.state.conditions.frames.inertial.time[:,0]
    noise_time        = np.linspace(time[0], time[-1], N) 
    noise_pos         = np.zeros((N,3)) 
    noise_pos[:,0]    = np.interp(noise_time,time,pos[:,0])
    noise_pos[:,1]    = np.interp(noise_time,time,pos[:,1])
    noise_pos[:,2]    = np.interp(noise_time,time,pos[:,2])
    
    num_gm_mic        = len(microphone_locations)  
    RML               = np.zeros((N,num_gm_mic,3)) 
    PHI               = np.zeros((N,num_gm_mic))
    THETA             = np.zeros((N,num_gm_mic)) 
    
    for cpt in range(N):  
        relative_locations         = np.zeros((num_gm_mic,3))
        relative_locations[:,0]    = microphone_locations[:,0] - (settings.aircraft_origin_location[0] + noise_pos[cpt,0])    
        relative_locations[:,1]    = microphone_locations[:,1] - (settings.aircraft_origin_location[1] + noise_pos[cpt,1]) 
        if MSL_altitude:
            relative_locations[:,2]    = -(noise_pos[cpt,2])  - microphone_locations[:,2] 
        else:
            relative_locations[:,2]    = -(noise_pos[cpt,2])
            
        RML[cpt,:,:]   = relative_locations 
        PHI[cpt,:]     =  np.arctan2(np.sqrt(np.square(relative_locations[:, 0]) + np.square(relative_locations[:, 1])),  relative_locations[:, 2])  
        THETA[cpt,:]   =  np.arctan2(relative_locations[:, 1], relative_locations[:, 0]) 
    
    return noise_time,noise_pos,RML,PHI,THETA,num_gm_mic 
 


# RCAIDE/Methods/Energy/Sources/Battery/Lithium_Ion_NMC/update_nmc_cell_age.py
# 
# 
# Created:  Feb 2024, M. Clarke 

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------
import numpy as np  
 
# ----------------------------------------------------------------------------------------------------------------------
# update_nmc_cell_age
# ----------------------------------------------------------------------------------------------------------------------  
def update_nmc_cell_age(battery,segment, battery_conditions,increment_battery_age_by_one_day):  
    """ This is an aging model for 18650 lithium-nickel-manganese-cobalt-oxide batteries. 
   
    Source: 
    Schmalstieg, Johannes, et al. "A holistic aging model for Li (NiMnCo) O2
    based 18650 lithium-ion batteries." Journal of Power Sources 257 (2014): 325-334.
      
    Assumptions:
    None

    Inputs:
      segment.conditions.energy. 
         battery.cycle_in_day                                                   [unitless]
         battery.cell.temperature                                               [Kelvin] 
         battery.voltage_open_circuit                                           [Volts] 
         battery.charge_throughput                                              [Amp-hrs] 
         battery.cell.state_of_charge                                           [unitless] 
    
    Outputs:
       segment.conditions.energy.
         battery.capacity_fade_factor     (internal resistance growth factor)   [unitless]
         battery.resistance_growth_factor (capactance (energy) growth factor)   [unitless]  
         
    Properties Used:
    N/A 
    """    
    n_series   = battery.electrical_configuration.series
    SOC        = battery_conditions.cell.state_of_charge
    V_ul       = battery_conditions.voltage_under_load/n_series
    t          = battery_conditions.cell.cycle_in_day         
    Q_prior    = battery_conditions.cell.charge_throughput[-1,0] 
    Temp       = np.mean(battery_conditions.cell.temperature) 
    
    # aging model  
    delta_DOD = abs(SOC[0][0] - SOC[-1][0])
    rms_V_ul  = np.sqrt(np.mean(V_ul**2)) 
    alpha_cap = (7.542*np.mean(V_ul) - 23.75) * 1E6 * np.exp(-6976/(Temp))  
    alpha_res = (5.270*np.mean(V_ul) - 16.32) * 1E5 * np.exp(-5986/(Temp))  
    beta_cap  = 7.348E-3 * (rms_V_ul - 3.667)**2 +  7.60E-4 + 4.081E-3*delta_DOD
    beta_res  = 2.153E-4 * (rms_V_ul - 3.725)**2 - 1.521E-5 + 2.798E-4*delta_DOD
    
    E_fade_factor   = 1 - alpha_cap*(t**0.75) - beta_cap*np.sqrt(Q_prior)   
    R_growth_factor = 1 + alpha_res*(t**0.75) + beta_res*Q_prior  
    
    battery_conditions.cell.capacity_fade_factor     = np.minimum(E_fade_factor,battery_conditions.cell.capacity_fade_factor)
    battery_conditions.cell.resistance_growth_factor = np.maximum(R_growth_factor,battery_conditions.cell.resistance_growth_factor)
    
    if increment_battery_age_by_one_day:
        battery_conditions.cell.cycle_in_day += 1 # update battery age by one day 
  
    return  



# RCAIDE/Methods/Energy/Sources/Battery/Common/compute_module_properties.py
# 
# 
# Created:  Jul 2023, M. Clarke 

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------

from RCAIDE.Framework.Core import Units 
import  numpy as  np

# ----------------------------------------------------------------------------------------------------------------------
#  METHOD
# ---------------------------------------------------------------------------------------------------------------------- -Common
def compute_module_properties(battery_module):  
    """Calculate module level properties of battery module using cell 
    properties and module configuraton
    
    Assumptions:
    Total battery module pack mass contains build-up factor (1.42) for battery module casing,
    internal wires, thermal management system and battery module management system 
    Factor computed using information of battery module properties for X-57 Maxwell 
    Aircraft
    
    Source:
    Cell Charge: Chin, J. C., Schnulo, S. L., Miller, T. B., Prokopius, K., and Gray, 
    J., Battery Performance Modeling on Maxwell X-57",AIAA Scitech, San Diego, CA,
    2019. URLhttp://openmdao.org/pubs/chin_battery_performance_x57_2019.pdf.     

    Inputs:
    mass              
    battery_module.cell
      nominal_capacity        [amp-hours]            
      nominal_voltage         [volts]
      pack_config             [unitless]
      mass                    [kilograms]
                          
    Outputs:              
     battery_module.             
       maximum_energy         [watt-hours]
       maximum_power              [watts]
       initial_maximum_energy [watt-hours]
       specific_energy        [watt-hours/kilogram]
       charging_voltage       [volts]
       mass_properties.    
        mass                  [kilograms] 
    """
    

    series_e           = battery_module.electrical_configuration.series
    parallel_e         = battery_module.electrical_configuration.parallel 
    normal_count       = battery_module.geometrtic_configuration.normal_count  
    parallel_count     = battery_module.geometrtic_configuration.parallel_count
    stacking_rows      = battery_module.geometrtic_configuration.stacking_rows

    if int(parallel_e*series_e) != int(normal_count*parallel_count):
        pass #raise Exception('Number of cells in gemetric layout not equal to number of cells in electric circuit configuration ')
        
        
    normal_spacing     = battery_module.geometrtic_configuration.normal_spacing   
    parallel_spacing   = battery_module.geometrtic_configuration.parallel_spacing
    volume_factor      = battery_module.volume_packaging_factor
    cell_diameter      = battery_module.cell.diameter
    cell_height        = battery_module.cell.height  
    euler_angles       = battery_module.orientation_euler_angles
    weight_factor      = battery_module.BMS_additional_weight_factor
    
    x1 =  normal_count * (cell_diameter + normal_spacing) * volume_factor # distance in the module-level normal direction
    x2 =  parallel_count * (cell_diameter + parallel_spacing) * volume_factor # distance in the module-level parallel direction
    x3 =  cell_height * volume_factor # distance in the module-level height direction 

    length = x1 / stacking_rows
    width  = x2
    height = x3 *stacking_rows     
    
    if  euler_angles[0] == (np.pi / 2):
        x1prime      = x2
        x2prime      = -x1
        x3prime      = x3 
    if euler_angles[1] == (np.pi / 2):
        x1primeprime = -x3prime
        x2primeprime = x2prime
        x3primeprime = x1prime
    if euler_angles[2] == (np.pi / 2):
        length       = x1primeprime
        width        = x3primeprime
        height       = -x2primeprime

    # store length, width and height
    battery_module.length = length
    battery_module.width  = width
    battery_module.height = height 
    
    amp_hour_rating                               = battery_module.cell.nominal_capacity   
    total_battery_assemply_mass                   = battery_module.cell.mass * series_e * parallel_e  
    battery_module.mass_properties.mass           = total_battery_assemply_mass*weight_factor  
    battery_module.specific_energy                = (amp_hour_rating*battery_module.cell.maximum_voltage)/battery_module.cell.mass  * Units.Wh/Units.kg
    battery_module.maximum_energy                 = total_battery_assemply_mass*battery_module.specific_energy    
    battery_module.specific_power                 = battery_module.specific_energy/battery_module.cell.nominal_capacity 
    battery_module.maximum_power                  = battery_module.specific_power*battery_module.mass_properties.mass  
    battery_module.maximum_voltage                = battery_module.cell.maximum_voltage  * series_e   
    battery_module.initial_maximum_energy         = battery_module.maximum_energy      
    battery_module.nominal_capacity               = battery_module.cell.nominal_capacity* parallel_e 
    battery_module.voltage                        = battery_module.maximum_voltage 



# RCAIDE/Library/Compoments/Airfoils/NACA_4_Series_Airfoil.py
# 
# 
# Created:  Mar 2024, M. Clarke 

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ---------------------------------------------------------------------------------------------------------------------- 
# RCAIDE imports     
from .Airfoil import Airfoil
 
# ---------------------------------------------------------------------------------------------------------------------- 
#  Airfoil
# ---------------------------------------------------------------------------------------------------------------------- 
 
class NACA_4_Series_Airfoil(Airfoil):
    """
    Class for generating NACA 4-series airfoil geometries.

    Attributes
    ----------
    tag : str
        Identifier for the airfoil (default: 'NACA_4_Series')
        
    NACA_4_Series_code : str
        Four-digit NACA designation (default: '0012')
        
    Notes
    -----
    Inherits all attributes from the base Airfoil class. The geometry is generated
    based on the NACA 4-series equations using the specified four-digit code.
    
    The NACA 4-series airfoils are defined by four digits that specify the camber,
    position of maximum camber, and thickness. For example, in a NACA 2412:
    - First digit (2) specifies maximum camber in percentage of chord
    - Second digit (4) specifies position of maximum camber in tenths of chord
    - Last two digits (12) specify maximum thickness in percentage of chord
    
    'NACA Code Format'
        MPXX where:
        - M is the maximum camber in percentage of chord
        - P is the position of maximum camber in tenths of chord
        - XX is the maximum thickness in percentage of chord
    
    See Also
    --------
    RCAIDE.Library.Components.Airfoils.Airfoil
        Base airfoil class
    """
    
    def __defaults__(self):
        """This sets the default values.
    
        Assumptions:
            None
        
        Source:
            None
        """        
        self.tag                   = 'NACA_4_Series'
        self.NACA_4_Series_code    = '0012'
        return 



# RCAIDE/Library/Components/Wings/__init__.py

"""
Module containing wing and lifting surface components for aircraft design and analysis. 
This module provides classes for various wing types including main wings, vertical and 
horizontal tails, and all-moving surfaces.

The module also includes a Control_Surfaces subpackage for various control surface types.

See Also
--------
RCAIDE.Library.Components.Airfoils
    Related module for airfoil definitions used in wing design
RCAIDE.Library.Components.Fuselages
    Related module for fuselage components that interface with wings
"""

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------

from .Segment                   import Segment, Segment_Container 


# RCAIDE/Library/Plots/Energy/__init__.py
# 

"""RCAIDE Package Setup
"""

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------

from .plot_solar_network_conditions             import plot_solar_network_conditions 
from .plot_battery_module_conditions            import plot_battery_module_conditions
from .plot_battery_cell_conditions              import plot_battery_cell_conditions
from .plot_battery_degradation                  import plot_battery_degradation 
from .plot_battery_temperature                  import plot_battery_temperature
from .plot_battery_module_C_rates               import plot_battery_module_C_rates 
from .plot_battery_ragone_diagram               import plot_battery_ragone_diagram 
from .plot_electric_propulsor_efficiencies      import plot_electric_propulsor_efficiencies
from .plot_fuel_consumption                     import plot_fuel_consumption
from .plot_altitude_sfc_weight                  import plot_altitude_sfc_weight
from .plot_propulsor_throttles                  import plot_propulsor_throttles               


# RCAIDE/Library/Plots/Thermal_Management/plot_cross_flow_heat_exchanger_conditions.py
# 
# 
# Created:  Sep 2024, S. Shekar

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------  

from RCAIDE.Framework.Core import Units
from RCAIDE.Library.Plots.Common import set_axes, plot_style
import matplotlib.pyplot as plt
import matplotlib.cm as cm
import numpy as np 

# ----------------------------------------------------------------------------------------------------------------------
#   plot_heat_exchanger_system_conditions
# ----------------------------------------------------------------------------------------------------------------------   
def plot_cross_flow_heat_exchanger_conditions(cross_flow_hex, results, coolant_line, save_figure,show_legend ,save_filename,file_type , width, height):
    """Plots the Cross Flow Heat Exchanger conditions throughout flight.
    
     Assumptions:
     None
    
     Source:
     None
    
     Inputs:
     results.segments.conditions.energy[coolant_line.tag][cross_flow_hex.tag].
                                                                             coolant_mass_flow_rate
                                                                             effectiveness_HEX   
                                                                             power          
                                                                             air_inlet_pressure
                                                                             inlet_air_temperature 
                                                                             air_mass_flow_rate    
                                                                       
     Outputs:
     Plots
    
     Properties Used:
     N/A	
     """ 
    
    # get plotting style 
    ps      = plot_style()  

    parameters = {'axes.labelsize': ps.axis_font_size,
                  'xtick.labelsize': ps.axis_font_size,
                  'ytick.labelsize': ps.axis_font_size,
                  'axes.titlesize': ps.title_font_size}
    plt.rcParams.update(parameters)
     
    # get line colors for plots 
    line_colors   = cm.inferno(np.linspace(0,0.9,len(results.segments)))     

    fig = plt.figure(save_filename)
    fig.set_size_inches(width,height)  
    axis_1 = plt.subplot(3,2,1)
    axis_2 = plt.subplot(3,2,2) 
    axis_3 = plt.subplot(3,2,3) 
    axis_4 = plt.subplot(3,2,4)
    axis_5 = plt.subplot(3,2,5) 
    axis_6 = plt.subplot(3,2,6) 
 
    for network in results.segments[0].analyses.energy.vehicle.networks: 
        busses  = network.busses 
        for bus in busses:
            for b_i, battery in enumerate(bus.battery_modules):
                if b_i == 0 or bus.identical_battery_modules == False:
                    for i in range(len(results.segments)): 
                        time    = results.segments[i].conditions.frames.inertial.time[:,0] / Units.min    
                        cross_flow_hex_conditions  = results.segments[i].conditions.energy[coolant_line.tag][cross_flow_hex.tag]  
                
                        coolant_mass_flow_rate     = cross_flow_hex_conditions.coolant_mass_flow_rate[:,0]        
                        effectiveness_HEX          = cross_flow_hex_conditions.effectiveness_HEX[:,0]   
                        power                      = cross_flow_hex_conditions.power[:,0]                       
                        inlet_air_pressure         = cross_flow_hex_conditions.air_inlet_pressure[:,0]          
                        inlet_air_temperature      = cross_flow_hex_conditions.inlet_air_temperature[:,0]          
                        air_mass_flow_rate         = cross_flow_hex_conditions.air_mass_flow_rate[:,0]     
                                             
                        if i == 0: 
                            axis_1.plot(time, effectiveness_HEX, color = line_colors[i], marker = ps.markers[b_i], linewidth = ps.line_width, label = cross_flow_hex.tag)
                        else:
                            axis_1.plot(time, effectiveness_HEX, color = line_colors[i], marker = ps.markers[b_i], linewidth = ps.line_width) 
                        axis_1.set_ylabel(r'Effectiveness') 
                        set_axes(axis_1)      
                
                        axis_2.plot(time,  inlet_air_temperature, color = line_colors[i], marker = ps.markers[b_i], linewidth = ps.line_width)
                        axis_2.set_ylabel(r'Air Temp. (K)') 
                        set_axes(axis_2)    
                        
                        axis_3.plot(time, coolant_mass_flow_rate, color = line_colors[i], marker = ps.markers[b_i], linewidth = ps.line_width)
                        axis_3.set_ylabel(r'Coolant $\dot{m}$ (kg/s)')
                        set_axes(axis_3) 
                
                        axis_4.plot(time, air_mass_flow_rate, color = line_colors[i], marker = ps.markers[b_i], linewidth = ps.line_width)
                        axis_4.set_ylabel(r'Air $\dot{m}$ (kg/s)')
                        set_axes(axis_4)                               
                 
                        axis_5.plot(time, power/1000, color = line_colors[i], marker = ps.markers[b_i], linewidth = ps.line_width)
                        axis_5.set_ylabel(r'HEX Power (KW)')
                        axis_5.set_xlabel(r'Time (mins)')
                        set_axes(axis_5)    
                
                        axis_6.plot(time, inlet_air_pressure/10e6 , color = line_colors[i], marker = ps.markers[b_i], linewidth = ps.line_width)
                        axis_6.set_ylabel(r'Air Pres. (MPa)')
                        axis_6.set_xlabel(r'Time (mins)')
                        set_axes(axis_6)  
            
    if show_legend:     
        leg =  fig.legend(bbox_to_anchor=(0.5, 0.95), loc='upper center', ncol = 4) 
        leg.set_title('Flight Segment', prop={'size': ps.legend_font_size, 'weight': 'heavy'})  
    
    # Adjusting the sub-plots for legend 
    fig.tight_layout()
    fig.subplots_adjust(top=0.8) 
    
    # set title of plot 
    title_text   = 'Heat_Exchanger_System'       
    fig.suptitle(title_text) 
    
    if save_figure:
        plt.savefig(save_filename + cross_flow_hex.tag + file_type)    
    return fig 